/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo;

import javax.annotation.Nonnull;
import javax.annotation.concurrent.Immutable;
import java.io.Serializable;
import org.openide.util.Lookup;
import it.tidalwave.util.As;
import it.tidalwave.geo.impl.AsSupport;

/***********************************************************************************************************************
 *
 * This class models a geographic coordinate composed of latitude, longitude and altitude. Angular quantities are
 * expressed in sexadecimal system, altitude in meters. Objects of this class are immutable.
 *
 * {@code Coordinate} is a {@link Lookup.Provider}, but the returned {@link Lookup} is always the same instance (this is
 * both for respecting the immutability constraint and for performance reasons). It is made for injecting stateless
 * services, such as finders etc.
 *
 * @author Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@Immutable
public class Coordinate extends AsSupport implements Serializable, Lookup.Provider, As
  {
    private static final long serialVersionUID = -403250971215465050L;

    public static final double EPSILON = 1E-6;

    private final double latitude;
    
    private final double longitude;
    
    private final double altitude;
    
    /*******************************************************************************************************************
     *
     * Creates a new {@code Coordinate} with given latitude, longitude and zero altitude.
     *
     * @param  latitude   the latitude
     * @param  longitude  the longitude
     *
     ******************************************************************************************************************/
    public Coordinate (final double latitude, final double longitude)
      {
        this(latitude, longitude, 0); 
      }
    
    /*******************************************************************************************************************
     *
     * Creates a new {@code Coordinate} with given latitude, longitude and altitude.
     *
     * @param  latitude   the latitude
     * @param  longitude  the longitude
     * @param  altitude   the altitude
     *
     ******************************************************************************************************************/
    public Coordinate (final double latitude, final double longitude, final double altitude)
      {
        this.latitude = latitude;
        this.longitude = longitude;
        this.altitude = altitude;
      }
    
    /*******************************************************************************************************************
     *
     * Returns the latitude.
     *
     * @return  latitude
     *
     ******************************************************************************************************************/
    public double getLatitude()
      {
        return latitude;
      }
    
    /*******************************************************************************************************************
     *
     * Returns the longitude.
     *
     * @return  longitude
     *
     ******************************************************************************************************************/
    public double getLongitude()
      {  
        return longitude;
      }
    
    /*******************************************************************************************************************
     *
     * Returns the altitude.
     *
     * @return  altitude
     *
     ******************************************************************************************************************/
    public double getAltitude()
      {  
        return altitude;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public boolean equals (final Object object)
      {
        if (object == null)
          {
            return false;
          }

        if (getClass() != object.getClass())
          {
            return false;
          }

        final Coordinate other = (Coordinate)object;

        return (Math.abs(this.latitude  - other.latitude)  < EPSILON) &&
               (Math.abs(this.longitude - other.longitude) < EPSILON) &&
               (Math.abs(this.altitude  - other.altitude)  < EPSILON);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public int hashCode()
      {
        int hash = 5;
        hash = 97 * hash + (int)(Double.doubleToLongBits(this.latitude) ^ (Double.doubleToLongBits(this.latitude) >>> 32));
        hash = 97 * hash + (int)(Double.doubleToLongBits(this.longitude) ^ (Double.doubleToLongBits(this.longitude) >>> 32));
        hash = 97 * hash + (int)(Double.doubleToLongBits(this.altitude) ^ (Double.doubleToLongBits(this.altitude) >>> 32));
        return hash;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    @Nonnull
    public String toString()
      {
        return CoordinateFormat.formatLatitude(latitude) + " " +
               CoordinateFormat.formatLongitude(longitude) + ", " +
                                altitude;
      }
  }
