/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.annotation.concurrent.Immutable;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.io.Serializable;
import org.openide.util.Lookup;
import it.tidalwave.util.As;
import it.tidalwave.geo.impl.AsSupport;
import static it.tidalwave.geo.Coordinate.EPSILON;

/***********************************************************************************************************************
 *
 * A {@code Sector} represents a geographical area comprised by two latitudes and longitudes.
 * 
 * @author Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@Immutable
public final class Sector extends AsSupport implements Serializable, Lookup.Provider, As
  {
    private static final long serialVersionUID = 3454353405934534L;

    private final double minLatitude;

    private final double maxLatitude;
    
    private final double minLongitude;

    private final double maxLongitude;

    private final static Coordinate ZERO = new Coordinate(0, 0);
    
    /*******************************************************************************************************************
     *
     * Creates an empty {@code Sector}.
     *
     ******************************************************************************************************************/
    public Sector()
      {
        minLatitude  = Double.NaN;
        maxLatitude  = Double.NaN;
        minLongitude = Double.NaN;
        maxLongitude = Double.NaN;
      }
    
    /*******************************************************************************************************************
     *
     * Creates a {@code Sector} from explicit parameters.
     *
     ******************************************************************************************************************/
    public Sector (final double minLatitude,
                   final double maxLatitude,
                   final double minLongitude,
                   final double maxLongitude)
      {
        this.minLatitude  = minLatitude;
        this.maxLatitude  = maxLatitude;
        this.minLongitude = minLongitude;
        this.maxLongitude = maxLongitude;
      }
    
    /*******************************************************************************************************************
     *
     * Creates a {@code Sector} as the smallest bounding box for the given set of coordinates.
     *
     ******************************************************************************************************************/
    public Sector (@Nonnull final Collection<Coordinate> coordinates)
      {
        double minLat = Double.MAX_VALUE;
        double maxLat = Double.MIN_VALUE;
        double minLon = Double.MAX_VALUE;
        double maxLon = Double.MIN_VALUE;

        for (final Coordinate c : coordinates)
          {
            final double latitude = c.getLatitude();
            final double longitude = c.getLongitude();
            minLat = Math.min(minLat, latitude);
            maxLat = Math.max(maxLat, latitude);
            minLon = Math.min(minLon, longitude);
            maxLon = Math.max(maxLon, longitude);
          }

        minLatitude = minLat;
        maxLatitude = maxLat;
        minLongitude = minLon;
        maxLongitude = maxLon;
      }

    /*******************************************************************************************************************
     *
     * Creates a {@code Sector} as the smallest bounding box for the given set of coordinates.
     *
     ******************************************************************************************************************/
    public Sector (@Nonnull final Coordinate ... coordinates)
      {
        this(Arrays.asList(coordinates));
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public double getMinLatitude()
      {
        return minLatitude;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public double getMinLongitude()
      {
        return minLongitude;
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public double getMaxLatitude()
      {
        return maxLatitude;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public double getMaxLongitude()
      {
        return maxLongitude;
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public double getLatitudeDelta()
      {
        return isVoid() ? 0.0 : (maxLatitude - minLatitude);
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public double getLongitudeDelta()
      {
        return isVoid() ? 0.0 : (maxLongitude - minLongitude);
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public double getMaxDelta()
      {
        return Math.max(getLatitudeDelta(), getLongitudeDelta());
      }
    
    /*******************************************************************************************************************
     *
     * Returns the center of this {@code Sector}. Void sectors have a conventional center at (0; 0).
     *
     * @return  the center 
     *
     ******************************************************************************************************************/
    @Nonnull 
    public Coordinate getCenter()
      {
        return isVoid() ? ZERO : new Coordinate((minLatitude + maxLatitude) / 2, (minLongitude + maxLongitude) / 2);
      }

    /*******************************************************************************************************************
     *
     * Returns the coordinate of the north-west corner of this  {@code Sector}. Void sectors have all corners
     * conventionally at (0; 0).
     *
     * @return  the coordinate of the north-west corner
     *
     ******************************************************************************************************************/
    @Nonnull
    public Coordinate getNWCorner()
      {
        return isVoid() ? ZERO : new Coordinate(maxLatitude, minLongitude);
      }

    /*******************************************************************************************************************
     *
     * Returns the coordinate of the north-east corner of this  {@code Sector}. Void sectors have all corners
     * conventionally at (0; 0).
     *
     * @return  the coordinate of the north-east corner
     *
     ******************************************************************************************************************/
    @Nonnull
    public Coordinate getNECorner()
      {
        return isVoid() ? ZERO : new Coordinate(maxLatitude, maxLongitude);
      }

    /*******************************************************************************************************************
     *
     * Returns the coordinate of the south-west corner of this  {@code Sector}. Void sectors have all corners
     * conventionally at (0; 0).
     *
     * @return  the coordinate of the south-west corner
     *
     ******************************************************************************************************************/
    @Nonnull
    public Coordinate getSWCorner()
      {
        return isVoid() ? ZERO : new Coordinate(minLatitude, minLongitude);
      }

    /*******************************************************************************************************************
     *
     * Returns the coordinate of the south-east corner of this  {@code Sector}. Void sectors have all corners
     * conventionally at (0; 0).
     *
     * @return  the coordinate of the south-east corner
     *
     ******************************************************************************************************************/
    @Nonnull
    public Coordinate getSECorner()
      {
        return isVoid() ? ZERO : new Coordinate(minLatitude, maxLongitude);
      }

    /*******************************************************************************************************************
     *
     * Returns a new {@code Sector} to fit an enclosed {@code Sector}.
     *
     ******************************************************************************************************************/
    @Nonnull
    public Sector with (@Nonnull final Sector inner)
      {
        if (inner.isVoid())
          {
            return this;
          }

        if (this.isVoid())
          {
            return inner;
          }

        return new Sector(Math.min(minLatitude, inner.minLatitude),
                          Math.max(maxLatitude, inner.maxLatitude),
                          Math.min(minLongitude, inner.minLongitude),
                          Math.max(maxLongitude, inner.maxLongitude));
      }
    
    /*******************************************************************************************************************
     *
     * Adjusts this {@code Sector} to fit the given {@link Coordinates}.
     *
     ******************************************************************************************************************/
    @Nonnull
    public Sector with (@Nonnull final Coordinate coordinate)
      {
        return with(Collections.singletonList(coordinate));
      }
    
    /*******************************************************************************************************************
     *
     * Adjusts this {@code Sector} to fit the given {@link Coordinates}.
     *
     ******************************************************************************************************************/
    @Nonnull
    public Sector with (@Nonnull final Collection<Coordinate> coordinates)
      {
        double minLat = isVoid() ? Double.MAX_VALUE : minLatitude;
        double maxLat = isVoid() ? Double.MIN_VALUE : maxLatitude;
        double minLon = isVoid() ? Double.MAX_VALUE : minLongitude;
        double maxLon = isVoid() ? Double.MIN_VALUE : maxLongitude;

        for (final Coordinate c : coordinates)
          {
            final double latitude = c.getLatitude();
            final double longitude = c.getLongitude();
            minLat = Math.min(minLat, latitude);
            maxLat = Math.max(maxLat, latitude);
            minLon = Math.min(minLon, longitude);
            maxLon = Math.max(maxLon, longitude);
          }

        return new Sector(minLat, maxLat, minLon, maxLon);
     }
    
    /*******************************************************************************************************************
     *
     * Returns true if this {@code Sector} contains another {@code Sector}.
     *
     ******************************************************************************************************************/
    public boolean contains (@Nonnull final Sector inner)
      {
        if (inner.isVoid())
          {
            return true;
          }

        if (this.isVoid())
          {
            return false;
          }

        return (this.minLatitude <= inner.minLatitude) &&
               (this.maxLatitude >= inner.maxLatitude) &&
               (this.minLongitude <= inner.minLongitude) &&
               (this.maxLongitude >= inner.maxLongitude);
      }


    /*******************************************************************************************************************
     *
     * Returns {@code true} if this {@code Sector} collapses on a single coordinate.
     *
     ******************************************************************************************************************/
    public boolean isSingleCoordinate()
      {
        return !isVoid() && (getMaxDelta() <= EPSILON);
      }

    /*******************************************************************************************************************
     *
     * Returns {@code true} if this {@code Sector} is void (it does not.designate any area).
     *
     ******************************************************************************************************************/
    public boolean isVoid()
      {
        return Double.isNaN(this.minLatitude);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public boolean equals (@CheckForNull final Object object)
      {
        if (object == null)
          {
            return false;
          }

        if (getClass() != object.getClass())
          {
            return false;
          }

        final Sector other = (Sector)object;

        if (this.isVoid() && other.isVoid())
          {
            return true;
          }

        if (this.isVoid() != other.isVoid())
          {
            return false;
          }

        return (Math.abs(this.minLatitude  - other.minLatitude)  <= EPSILON) &&
               (Math.abs(this.maxLatitude  - other.maxLatitude)  <= EPSILON) &&
               (Math.abs(this.minLongitude - other.minLongitude) <= EPSILON) &&
               (Math.abs(this.maxLongitude - other.maxLongitude) <= EPSILON);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public int hashCode()
      {
        int hash = 7;
        hash = 79 * hash + (int)(Double.doubleToLongBits(this.minLatitude) ^ (Double.doubleToLongBits(this.minLatitude) >>> 32));
        hash = 79 * hash + (int)(Double.doubleToLongBits(this.maxLatitude) ^ (Double.doubleToLongBits(this.maxLatitude) >>> 32));
        hash = 79 * hash + (int)(Double.doubleToLongBits(this.minLongitude) ^ (Double.doubleToLongBits(this.minLongitude) >>> 32));
        hash = 79 * hash + (int)(Double.doubleToLongBits(this.maxLongitude) ^ (Double.doubleToLongBits(this.maxLongitude) >>> 32));
        return hash;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    @Nonnull 
    public String toString()
      {
        return isVoid() ? "Sector[void]"
                        : String.format("Sector[%s -> %s, %s -> %s]",
                          CoordinateFormat.formatLatitude(maxLatitude),
                          CoordinateFormat.formatLatitude(minLatitude),
                          CoordinateFormat.formatLongitude(minLongitude),
                          CoordinateFormat.formatLongitude(maxLongitude));
      }
  }
