/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo;

import org.junit.Test;
import static org.junit.Assert.*;
import static org.hamcrest.CoreMatchers.*;
import static org.hamcrest.number.IsCloseTo.*;
import static it.tidalwave.geo.Coordinate.EPSILON;

/***********************************************************************************************************************
 *
 * @author Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class SectorTest
  {
    @Test
    public void defaultSectorShouldBeVoid()
      {
        final Sector voidSector = new Sector();
        assertThat(voidSector.isSingleCoordinate(), is(false));
        assertThat(voidSector.isVoid(), is(true));
        assertThat(voidSector.getMinLatitude(), is(Double.NaN));
        assertThat(voidSector.getMaxLatitude(), is(Double.NaN));
        assertThat(voidSector.getMinLongitude(), is(Double.NaN));
        assertThat(voidSector.getMaxLongitude(), is(Double.NaN));
        assertThat(voidSector.getLatitudeDelta(), is(0.0));
        assertThat(voidSector.getLongitudeDelta(), is(0.0));
        assertThat(voidSector.getMaxDelta(), is(0.0));
        assertThat(voidSector.getCenter(), is(new Coordinate(0.0, 0.0, 0.0)));
        assertThat(voidSector.getNWCorner(), is(new Coordinate(0, 0, 0)));
        assertThat(voidSector.getNECorner(), is(new Coordinate(0, 0, 0)));
        assertThat(voidSector.getSWCorner(), is(new Coordinate(0, 0, 0)));
        assertThat(voidSector.getSECorner(), is(new Coordinate(0, 0, 0)));
        assertThat(voidSector.contains(voidSector), is(true));
        assertThat(voidSector.toString(), is("Sector[void]"));
      }

    @Test
    public void voidSectorsShouldBeMutuallyEquals()
      {
        final Sector void1 = new Sector();
        final Sector void2 = new Sector();
        assertThat(void1.equals(void2), is(true));
        assertThat(void2.equals(void1), is(true));
      }

    @Test
    public void voidSectorsShouldHaveSameHashcode()
      {
        final Sector void1 = new Sector();
        final Sector void2 = new Sector();
        assertThat(void1.hashCode(), is(void2.hashCode()));
      }

    @Test
    public void testSingleCoordinateSector()
      {
        final Sector sector = new Sector(new Coordinate(10.0, 20.0, 30.0));
        assertThat(sector.isSingleCoordinate(), is(true));
        assertThat(sector.isVoid(), is(false));
        assertThat(sector.getMinLatitude(), is(closeTo(10.0, EPSILON)));
        assertThat(sector.getMaxLatitude(), is(closeTo(10.0, EPSILON)));
        assertThat(sector.getMinLongitude(), is(closeTo(20.0, EPSILON)));
        assertThat(sector.getMaxLongitude(), is(closeTo(20.0, EPSILON)));
        assertThat(sector.getLatitudeDelta(), is(closeTo(0.0, EPSILON)));
        assertThat(sector.getLongitudeDelta(), is(closeTo(0.0, EPSILON)));
        assertThat(sector.getMaxDelta(), is(closeTo(0.0, EPSILON)));
        assertThat(sector.getCenter(), is(new Coordinate(10.0, 20.0, 0.0)));
        assertThat(sector.getNWCorner(), is(new Coordinate(10, 20, 0)));
        assertThat(sector.getNECorner(), is(new Coordinate(10, 20, 0)));
        assertThat(sector.getSWCorner(), is(new Coordinate(10, 20, 0)));
        assertThat(sector.getSECorner(), is(new Coordinate(10, 20, 0)));
        assertThat(sector.contains(sector), is(true));
        assertThat(sector.toString(), is("Sector[N 10\u00b0 00' 00.000\" -> N 10\u00b0 00' 00.000\", " +
                                                "E 20\u00b0 00' 00.000\" -> E 20\u00b0 00' 00.000\"]"));
      }

    @Test
    public void testTwoCoordinatesSector()
      {
        final Sector sector = new Sector(new Coordinate(10.0, 20.0, 0.0), new Coordinate(-30.0, -50.0, 1000.0));
        assertThat(sector.isSingleCoordinate(), is(false));
        assertThat(sector.isVoid(), is(false));
        assertThat(sector.getMinLatitude(), is(closeTo(-30.0, EPSILON)));
        assertThat(sector.getMaxLatitude(), is(closeTo(10.0, EPSILON)));
        assertThat(sector.getMinLongitude(), is(closeTo(-50.0, EPSILON)));
        assertThat(sector.getMaxLongitude(), is(closeTo(20.0, EPSILON)));
        assertThat(sector.getLatitudeDelta(), is(closeTo(40.0, EPSILON)));
        assertThat(sector.getLongitudeDelta(), is(closeTo(70.0, EPSILON)));
        assertThat(sector.getMaxDelta(), is(closeTo(70.0, EPSILON)));
        assertThat(sector.getCenter(), is(new Coordinate(-10.0, -15.0, 0.0)));
        assertThat(sector.getNWCorner(), is(new Coordinate(10, -50, 0)));
        assertThat(sector.getNECorner(), is(new Coordinate(10, 20, 0)));
        assertThat(sector.getSWCorner(), is(new Coordinate(-30, -50, 0)));
        assertThat(sector.getSECorner(), is(new Coordinate(-30, 20, 0)));
        assertThat(sector.contains(sector), is(true));
        assertThat(sector.toString(), is("Sector[N 10\u00b0 00' 00.000\" -> S 30\u00b0 00' 00.000\", " +
                                                "W 50\u00b0 00' 00.000\" -> E 20\u00b0 00' 00.000\"]"));
      }

    @Test
    public void testMultipleCoordinatesSector()
      {
        final Coordinate[] coordinates = new Coordinate[]
          {
            new Coordinate( 10.0,   20.0, 110.0),
            new Coordinate( 54.0, -120.0, 340.0),
            new Coordinate(-43.0,  150.0, 450.0),
            new Coordinate(-34.0,  -10.0, -30.0),
            new Coordinate(-30.0,  -90.0, 465.0),
            new Coordinate( 16.0,  -75.0, 234.0),
            new Coordinate( 26.0,  100.0, 789.0),
            new Coordinate(-37.0,  101.0,  43.0),
            new Coordinate(-56.0,  -40.0, 903.0)
          };
        final Sector sector = new Sector(coordinates);
        assertThat(sector.isSingleCoordinate(), is(false));
        assertThat(sector.isVoid(), is(false));
        assertThat(sector.getMinLatitude(), is(closeTo(-56.0, EPSILON)));
        assertThat(sector.getMaxLatitude(), is(closeTo(54.0, EPSILON)));
        assertThat(sector.getMinLongitude(), is(closeTo(-120.0, EPSILON)));
        assertThat(sector.getMaxLongitude(), is(closeTo(150.0, EPSILON)));
        assertThat(sector.getLatitudeDelta(), is(closeTo(110.0, EPSILON)));
        assertThat(sector.getLongitudeDelta(), is(closeTo(270.0, EPSILON)));
        assertThat(sector.getMaxDelta(), is(closeTo(270.0, EPSILON)));
        assertThat(sector.getCenter(), is(new Coordinate(-1.0, 15.0, 0.0)));
        assertThat(sector.getNWCorner(), is(new Coordinate(54, -120, 0)));
        assertThat(sector.getNECorner(), is(new Coordinate(54, 150, 0)));
        assertThat(sector.getSWCorner(), is(new Coordinate(-56, -120, 0)));
        assertThat(sector.getSECorner(), is(new Coordinate(-56, 150, 0)));
        assertThat(sector.contains(sector), is(true));
        assertThat(sector.toString(), is("Sector[N 54\u00b0 00' 00.000\" -> S 56\u00b0 00' 00.000\", " +
                                                "W 120\u00b0 00' 00.000\" -> E 150\u00b0 00' 00.000\"]"));
      }
  }