/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.geocoding.impl;

import javax.annotation.Nonnull;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.SortedSet;
import javax.xml.namespace.QName;
import java.io.Serializable;
import org.openide.util.Lookup;
import it.tidalwave.util.AsException;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.util.spi.SimpleFinderSupport;
import it.tidalwave.role.LocalizedDisplayable;
import it.tidalwave.geo.Coordinate;
import it.tidalwave.geo.geocoding.GeoCoder;
import it.tidalwave.geo.geocoding.GeoCoderEntity;
import org.junit.Before;
import org.junit.Test;
import org.mockito.invocation.InvocationOnMock;
import org.mockito.stubbing.Answer;
import static it.tidalwave.role.LocalizedDisplayable.LocalizedDisplayable;
import static org.junit.Assert.*;
import static org.mockito.Mockito.*;
import static org.hamcrest.CoreMatchers.*;

/***********************************************************************************************************************
 *
 * @author Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class DefaultGeoCoderEntityTest
  {
    public static class Capability implements Serializable
      {
      }

    private final static String GEO_CODER_ID = "http://geocoder";
    private final static String ID = "theId";
    private final static String NAME = "theName";
    private final static String CODE = "theCode";
    private final static String TYPE_AS_STRING = "theType";
    private final static Coordinate COORDINATE = new Coordinate(1, 2, 3);
    private final static Capability CAP1 = new Capability();
    private final static Capability CAP2 = new Capability();

    private GeoCoder geoCoder;
    private DefaultGeoCoderEntity fixture;
    private GeoCoderEntity parent;
    private GeoCoderEntity child1;
    private GeoCoderEntity child2;
    private GeoCoderEntity child3;

    @Before
    public void setup()
       throws NotFoundException
      {
        geoCoder = mock(GeoCoder.class);
        parent = mock(GeoCoderEntity.class);
        child1 = mock(GeoCoderEntity.class);
        child2 = mock(GeoCoderEntity.class);
        child3 = mock(GeoCoderEntity.class);

        when(geoCoder.getId()).thenReturn(GEO_CODER_ID);
        when(geoCoder.findGeoEntityById(eq(ID))).thenAnswer(new Answer<GeoCoderEntity>()
          {
            @Override @Nonnull
            public GeoCoderEntity answer (final @Nonnull InvocationOnMock invocation)
              {
                fixture.parent = parent;
                return fixture;
              }
          });

        when(geoCoder.typeFromString(eq(TYPE_AS_STRING))).thenReturn(GeoCoderEntity.Type.ADMIN_DIVISION_3);
        when(geoCoder.findChildren(eq(ID))).thenReturn(new SimpleFinderSupport<GeoCoderEntity>("mock")
          {
            @Override @Nonnull
            protected List<GeoCoderEntity> computeResults()
              {
                return Arrays.asList(child1, child2, child3);
              }
          });

        fixture = new DefaultGeoCoderEntity(geoCoder, parent, ID, NAME, COORDINATE, CODE, TYPE_AS_STRING, new Serializable[]{CAP1, CAP2});
      }

    @Test
    public void shouldReturnTheCorrectGeoCoder()
      {
        assertThat(fixture.getGeoCoder(), is(sameInstance(geoCoder)));
      }

    @Test
    public void shouldReturnTheCorrectId()
      {
        assertThat(fixture.getId(), is(ID));
      }

    @Test
    public void shouldReturnTheCorrectCoordinate()
      {
        assertThat(fixture.getCoordinate(), is(sameInstance(COORDINATE)));
      }

    @Test
    public void shouldReturnTheCorrectCode()
      {
        assertThat(fixture.getCode(), is(CODE));
      }

    @Test
    public void shouldReturnTheCorrectTypeAsString()
      {
        assertThat(fixture.getTypeAsString(), is(TYPE_AS_STRING));
      }

    @Test
    public void shouldReturnTheCorrectType()
      {
        assertThat(fixture.getType(), is(GeoCoderEntity.Type.ADMIN_DIVISION_3));
        verify(geoCoder).typeFromString(eq(TYPE_AS_STRING));
      }

    @Test
    public void shouldReturnTheCorrectChildrenAndGeoCoderIsOnlyCalledOnce()
      {
        final List<GeoCoderEntity> children = Arrays.asList(child1, child2, child3);
        assertThat((List<GeoCoderEntity>)fixture.findChildren().results(), is(children));
        assertThat((List<GeoCoderEntity>)fixture.findChildren().results(), is(children));
        assertThat((List<GeoCoderEntity>)fixture.findChildren().results(), is(children));
        assertThat((List<GeoCoderEntity>)fixture.findChildren().results(), is(children));
        verify(geoCoder, times(1)).findChildren(eq(ID));
      }

    @Test
    public void shouldReturnTheCorrectParent()
      throws NotFoundException
      {
        assertThat(fixture.getParent(), is(sameInstance(parent)));
      }

    @Test
    public void shouldReturnTheCorrectParentAfterIsNulledAndGeoCoderIsOnlyCalledOnce()
      throws NotFoundException
      {
        assertThat(fixture.getParent(), is(sameInstance(parent)));
        fixture.parent = null;
        assertThat(fixture.getParent(), is(sameInstance(parent)));
        assertThat(fixture.getParent(), is(sameInstance(parent)));
        assertThat(fixture.getParent(), is(sameInstance(parent)));
        verify(geoCoder, times(1)).findGeoEntityById(ID);
      }

    @Test
    public void shouldActAsACorrectDisplayable()
      {
        final LocalizedDisplayable displayable = fixture.as(LocalizedDisplayable);
        assertThat(displayable, is(notNullValue()));
        assertThat(displayable.getDisplayName(), is(NAME));
        assertThat(displayable.getDisplayName(Locale.ENGLISH), is(NAME));

        final Map<Locale, String> displayNames = displayable.getDisplayNames();
        assertThat(displayNames, is(notNullValue()));
        assertThat(displayNames.size(), is(1));
        final Entry<Locale, String> entry = displayNames.entrySet().iterator().next();
        assertThat(entry.getKey(), is(Locale.ENGLISH));
        assertThat(entry.getValue(), is(NAME));

        final SortedSet<Locale> locales = displayable.getLocales();
        assertThat(locales, is(notNullValue()));
        assertThat(locales.size(), is(1));
        assertThat(locales.first(), is(Locale.ENGLISH));
      }

    @Test
    public void shouldCreateLookupOnlyOnce()
      {
        final Lookup lookup1 = fixture.getLookup();
        assertThat(lookup1, is(not(nullValue())));
        final Lookup lookup2 = fixture.getLookup();
        assertThat(lookup2, is(not(nullValue())));
        assertThat(lookup1, is(sameInstance(lookup2)));
      }

    @Test
    public void shouldReCreateLookupAfterIsNulled()
      {
        final Lookup lookup1 = fixture.getLookup();
        assertThat(lookup1, is(not(nullValue())));
        fixture.lookup = null;
        final Lookup lookup2 = fixture.getLookup();
        assertThat(lookup2, is(not(nullValue())));
        assertThat(lookup1, is(not(sameInstance(lookup2))));
      }

    @Test
    public void shouldLookupTheCapabilities()
      {
        final Collection<? extends Capability> capabilities = fixture.getLookup().lookupAll(Capability.class);
        assertThat(capabilities.size(), is(2));
        assertThat(capabilities.contains(CAP1), is(true));
        assertThat(capabilities.contains(CAP2), is(true));
      }

    @Test
    public void shouldLookupTheCoordinate()
      {
        final Coordinate coordinate = fixture.getLookup().lookup(Coordinate.class);
        assertThat(coordinate, is(sameInstance(COORDINATE)));
      }

    @Test
    public void shouldLookupTheQName()
      {
        final QName qName = fixture.getLookup().lookup(QName.class);
        assertThat(qName, is(new QName(GEO_CODER_ID + ID + "/")));
      }

    @Test
    public void shouldWorkAsCapabilities()
      {
        final Capability as = fixture.as(Capability.class);
        assertThat(as, is(notNullValue()));
      }

    @Test(expected=AsException.class)
    public void shouldThrowExceptionWhenAsWithWrongPersonality()
      {
        fixture.as(Date.class);
      }

    @Test
    public void shouldReturnTheCorrectToString()
      {
        assertThat(fixture.toString(), is(String.format("DefaultGeoCoderEntity@%x[%s, %s]",
                                                        System.identityHashCode(fixture), ID, NAME)));
      }
  }
