/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
/***********************************************************************************************************************
 *
 * <p>This package provides a visual component for navigating inside a hierarchy of geographical entities, rendered as
 * a tree.
 *
 * <!-- img src="GeoExplorerPresentation.png"/ -->
 *
 * This component does not expose a single interface for services; instead, it works as a factory where specific
 * services can be looked up:
 *
 * <pre>
 * DataLoader dataLoader = geoExplorerPresentation.getLookup().lookup(DataLoader.class);
 * dataLoader.loadData();
 * </pre>
 *
 * The {@link it.tidalwave.netbeans.role.DataLoader DataLoader} can be used to force the reloading of data in the
 * tree.</p>
 *
 *
 * <h3>Injectable roles</h3>
 *
 * <p>The implementation of this component is delegated to a set of small and cohese classes named "roles". Each
 * one has got a single reponsibility and usually can be replaced independently of the others. Each role has a
 * symbolic name, which is part of the public API, and an implementation class.<br/>
 * Most roles have a default implementation; some are mandatory (i.e. their absence breaks the
 * {@code GeoExplorerPresentation}}, while others are optional and can be added by the programmer.</p>
 *
 * <table border="1" cellpadding="4" cellspacing="0">
 * <tr><th>Role</th><th>Description</th><th>Implementation class</th><th>Description</th></tr>
 * <tr>
 * <td rowspan="2">{@code DataLoader}<br/>(mandatory)</td>
 * <td rowspan="2">Controls <i>where</i> data for populating the tree are loaded from.</td>
 * <td>default</td>
 * <td>Loads a complete hierarchy from the default {@link GeoCoder}.</td>
 * </tr>
 * <tr>
 * <td>custom roles</td>
 * <td>Can be implemented by subclassing {@link it.tidalwave.netbeans.windows.role.NodeDataLoader NodeDataLoader}.
 * </td>
 * </tr>
 *
 * <tr>
 * <td rowspan="3">{@code DataLoaderStrategy}<br/>(mandatory)</td>
 * <td rowspan="3">Controls <i>when</i>, in the life-cycle of the component, data are loaded.</td>
 * <td>{@link it.tidalwave.netbeans.windows.role.LazyDataLoaderStrategy LazyDataLoaderStrategy}<br/>default</td>
 * <td>Loads data the first time the component is shown on the screen. If the operation requires a long time,
 *     a "please wait" message is shown on the tree.</td>
 * </tr>
 * <tr>
 * <td>{@link it.tidalwave.netbeans.windows.role.EagerDataLoaderStrategy EagerDataLoaderStrategy}</td>
 * <td>Loads data when the component is created.</td>
 * </tr>
 * <tr>
 * <td>custom roles</td>
 * <td>Can be implemented by subclassing
 *     {@link it.tidalwave.netbeans.windows.role.TopComponentDataLoaderStrategy TopComponentDataLoaderStrategy}.
 * </td>
 * </tr>
 *
 * <tr>
 * <td rowspan="2">{@code ExplorerPanel}<br/>(mandatory)</td>
 * <td rowspan="2">A visual container of the rendering component.</td>
 * <td>{@code ExplorerPanel}<br/>default</td>
 * <td>Eventually applies transition effects when data are reloaded.</td>
 * </tr>
 * <tr>
 * <td>custom roles</td>
 * <td>Usually there should be no reasons for replace it, because the actual transition is managed by a
 *     {@code TransitionController} (see below). Can be implemented by subclassing
 *     {@link it.tidalwave.netbeans.explorer.ExplorerPanel ExplorerPanel}.</td>
 * </tr>
 *
 * <tr>
 * <td rowspan="2">{@code GUIBuilder}<br/>(mandatory)</td>
 * <td rowspan="2">The builder that creates the UI for this component.</td>
 * <td>{@code SingleExplorerGUIBuilder}<br/>default</td>
 * <td>This builder makes most of the available room for a node explorer, with a button toolbar at the topmost part.</td>
 * </tr>
 * <tr>
 * <td>custom roles</td>
 * <td>Can be implemented by subclassing
 *     {@link it.tidalwave.netbeans.windows.GUIBuilder GUIBuilder}.</td>
 * </tr>
 *
 * <tr>
 * <td>{@code NodeActivator}<br/>(optional)</td>
 * <td>Implements the standard role for activating {@code Node}s in a {@code TopComponent}.</td>
 * <td>{@link it.tidalwave.netbeans.windows.role.NodeActivator NodeActivator}<br/>default</td>
 * <td>Activates the selected node in the {@link TopComponent}. This is the usual role for a NetBeans Platform
 *     application and it is required for enabling contextual actions and similar stuff.</td>
 * </tr>
 * <tr>
 *
 * <tr>
 * <td>{@code NodeDecorator}<br/>(optional)</td>
 * <td>Allows to decorate the {@code Node}s independently from their origin.</td>
 * <td>custom roles</td>
 * <td>Can be implemented by subclassing {@link it.tidalwave.netbeans.nodes.NodeDecorator NodeDecorator}.
 * Multiple decorators can be injected at the same time.
 * </td>
 * </tr>
 * <tr>
 *
 * <tr>
 * <td rowspan="2">{@code NodeView}<br/>(mandatory)</td>
 * <td rowspan="2">The component that actually renders the {@code Node}s.</td>
 * <td>{@link it.tidalwave.netbeans.explorer.view.EnhancedBeanTreeView EnhancedBeanTreeView}<br/>default</td>
 * <td>Renders {@code Node}s in a simple tree fashion.</td>
 * </tr>
 * <tr>
 * <td>custom roles</td>
 * <td>Any class that is able to work with an {@link org.openide.explorer.ExplorerManager ExplorerManager}.</td>
 * </tr>
 *
 * <tr>
 * <td rowspan="2">{@code SelectionHandler}<br/>(mandatory)</td>
 * <td rowspan="2">Defines what to do when a selection is performed in the tree.</td>
 * <td>default</td>
 * <td>Publishes selected {@link it.tidalwave.geo.geocoding.GeoEntity GeoEntity} instances to an
 * {@link org.netbeans.platformx.eventbus.api.EventBus EventBus}, both as a
 * {@link it.tidalwave.geo.geocoding.GeoEntitySet GeoEntitySet} and the enclosing
 * {@link it.tidalwave.geo.Sector Sector}.</td>
 * </tr>
 * <tr>
 * <td>custom roles</td>
 * <td>Can be implemented by subclassing
 *     {@link it.tidalwave.geo.explorer.role.SelectionHandler SelectionHandler}.</td>
 * </tr>
 *
 * <tr>
 * <td rowspan="2">{@code SelectionStrategy}<br/>(mandatory)</td>
 * <td rowspan="2">Defines the way selections are performed in the tree.</td>
 * <td>default</td>
 * <td>Automatically performs a selection when a single click is done on a node.</td>
 * </tr>
 * <tr>
 * <td>custom roles</td>
 * <td>Can be implemented by subclassing
 *     {@link it.tidalwave.netbeans.explorer.role.ExplorerSelectionStrategy ExplorerSelectionStrategy}.</td>
 * </tr>
 *
 * <tr>
 * <td rowspan="2">{@code TransitionController}<br/>(optional)</td>
 * <td rowspan="2">Defines the graphic transition to implement when data are reloaded.</td>
 * <td>none by default</td>
 * <td>Applies a graphic transition when data are reloaded.</td>
 * </tr>
 * <tr>
 * <td>custom roles</td>
 * <td>Can be implemented by subclassing
 *     {@link TransitionController}.</td>
 * </tr>
 * </table>
 * <br/>
 *
 *
 * <h3>Customization and configuration</h3>
 *
 * Configuration is performed by means of {@code layer.xml}. The following is the default configuration of
 * {@code GeoExplorerPresentation}:
 *
 * <pre>
 * &lt;?xml version="1.0" encoding="UTF-8"?&gt;
 * &lt;!DOCTYPE filesystem PUBLIC "-//NetBeans//DTD Filesystem 1.1//EN" "http://www.netbeans.org/dtds/filesystem-1_1.dtd"&gt;
 * &lt;filesystem&gt;
 *     &lt;folder name="Roles"&gt;
 *         &lt;folder name="it.tidalwave.geo.explorer.GeoExplorerPresentation"&gt;
 *             &lt;file name="DataLoader.instance"&gt;
 *                 ...
 *             &lt;/file&gt;
 *             &lt;file name="DataLoaderStrategy.instance"&gt;
 *                 ...
 *             &lt;/file&gt;
 *             &lt;file name="ExplorerPanel.instance"&gt;
 *                 ...
 *             &lt;/file&gt;
 *             &lt;file name="GUIBuilder.instance"&gt;
 *                 ...
 *             &lt;/file&gt;
 *             &lt;file name="NodeActivator.instance"&gt;
 *                 ...
 *             &lt;/file&gt;
 *             &lt;file name="NodeView.instance"&gt;
 *                 ...
 *             &lt;/file&gt;
 *             &lt;file name="SelectionStrategy.instance"&gt;
 *                 ...
 *             &lt;/file&gt;
 *             &lt;file name="SelectionHandler.instance"&gt;
 *                 ...
 *             &lt;/file&gt;
 *         &lt;/folder&gt;
 *     &lt;/folder&gt;
 * ...
 * &lt;/filesystem&gt;
 * </pre>
 *
 * In order to configure {@code GeoExplorerPresentation} you need to create a specific NetBeans Platform module with
 * a dependency on this module and a {@code layer.xml}.
 * To replace a default role, the relevant definition must be overridden. For instance, this configuration replaces
 * the "lazy data loader" strategy with the "eager" one:
 *
 * <pre>
 * &lt;?xml version="1.0" encoding="UTF-8"?&gt;
 * &lt;!DOCTYPE filesystem PUBLIC "-//NetBeans//DTD Filesystem 1.1//EN" "http://www.netbeans.org/dtds/filesystem-1_1.dtd"&gt;
 * &lt;filesystem&gt;
 *     &lt;folder name="Roles"&gt;
 *         &lt;folder name="it.tidalwave.geo.explorer.GeoExplorerPresentation"&gt;
 *             <b>&lt;file name="DataLoaderStrategy.instance"&gt;
 *                 &lt;attr name="instanceClass" stringvalue="it.tidalwave.netbeans.windows.role.EagerDataLoaderStrategy"/&gt;
 *             &lt;/file&gt;</b>
 *         &lt;/folder&gt;
 *     &lt;/folder&gt;
 * &lt;/filesystem&gt;
 * </pre>
 *
 * It is also possible to disable a default role without replacing it; the following configuration disables the
 * "node activator":
 *
 * <pre>
 * &lt;?xml version="1.0" encoding="UTF-8"?&gt;
 * &lt;!DOCTYPE filesystem PUBLIC "-//NetBeans//DTD Filesystem 1.1//EN" "http://www.netbeans.org/dtds/filesystem-1_1.dtd"&gt;
 * &lt;filesystem&gt;
 *     &lt;folder name="Roles"&gt;
 *         &lt;folder name="it.tidalwave.geo.explorer.GeoExplorerPresentation"&gt;
 *             <b>&lt;file name="NodeActivator.instance_hidden"/&gt;</b>
 *         &lt;/folder&gt;
 *     &lt;/folder&gt;
 * &lt;/filesystem&gt;
 * </pre>
 *
 * Last but not least, it is also possible to add new roles:
 *
 * <pre>
 * &lt;?xml version="1.0" encoding="UTF-8"?&gt;
 * &lt;!DOCTYPE filesystem PUBLIC "-//NetBeans//DTD Filesystem 1.1//EN" "http://www.netbeans.org/dtds/filesystem-1_1.dtd"&gt;
 * &lt;filesystem&gt;
 *     &lt;folder name="Roles"&gt;
 *         &lt;folder name="it.tidalwave.geo.explorer.GeoExplorerPresentation"&gt;
 *             <b>&lt;file name="MyNewRole.instance"&gt;
 *                 &lt;attr name="instanceClass" stringvalue="my.newrole.Implementation"/&gt;
 *             &lt;/file&gt;</b>
 *         &lt;/folder&gt;
 *     &lt;/folder&gt;
 * &lt;/filesystem&gt;
 * </pre>
 *
 * Have a look at {@link it.tidalwave.netbeans.rule.util.BeanFactory} for a convenient way to instantiate beans by means
 * of {@code layer.xml} also populating their properties.
 *
 *
 * <h3>Default state</h3>
 *
 * This component is placed by default in the "explorer" mode and it's not set visible by default. If you want to be
 * visible by default, use this customization fragment of {@code layer.xml}:
 *
 * <pre>
 * &lt;?xml version="1.0" encoding="UTF-8"?&gt;
 * &lt;!DOCTYPE filesystem PUBLIC "-//NetBeans//DTD Filesystem 1.1//EN" "http://www.netbeans.org/dtds/filesystem-1_1.dtd"&gt;
 * &lt;filesystem&gt;
 *     &lt;folder name="Windows2"&gt;
 *         &lt;folder name="Modes"&gt;
 *             &lt;folder name="explorer"&gt;
 *                 &lt;file name="it.tidalwave.geo.explorer.GeoExplorerPresentation.wstcref" url="GeoExplorerTopComponentWstcref.xml"/&gt;
 *             &lt;/folder&gt;
 *         &lt;/folder&gt;
 *     &lt;/folder&gt;
 * &lt;/filesystem&gt;
 * </pre>
 *
 * with {@code GeoExplorerTopComponentWstcref.xml} being:
 *
 * <pre>
 * &lt;?xml version="1.0" encoding="UTF-8"?&gt;
 * &lt;!DOCTYPE tc-ref PUBLIC "-//NetBeans//DTD Top Component in Mode Properties 2.0//EN" "http://www.netbeans.org/dtds/tc-ref2_0.dtd"&gt;
 * &lt;tc-ref version="2.0" &gt;
 *     &lt;module name="it.tidalwave.geo.explorer" spec="1.0"/&gt;
 *     &lt;tc-id id="it.tidalwave.geo.explorer.GeoExplorerPresentation"/&gt;
 *     &lt;state opened="true"/&gt;
 * &lt;/tc-ref&gt;
 * </pre>
 *
 *
 * <h3>Predefined actions</h3>
 *
 * An action to toggle the visibility status of the {@link GeoExplorerPresentation} is automatically installed under the
 * "Window" menu. The name of the file representing the action, that can be used to customize the position or remove the
 * action completely, is
 *
 * <pre>
 * "Actions/Window/it-tidalwave-geo-explorer-impl-action-GeoExplorerAction.instance"
 * </pre>
 *
 *
 * <h3>Node rendering and other properties</h3>
 *
 * This component is not responsible for the contents of the tree and for the rendering of nodes; these semantics are
 * managed by a {@code NodeManager}. If you don't like the default, you can customize the contents by using a
 * {@link org.openide.nodes.FilterNode} injected by a {@link it.tidalwave.netbeans.nodes.NodeDecorator NodeDecorator}
 * or completely replace the contents by using a different {@code DataLoader}.
 *
 * TBD: add example of a decorator
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.explorer;
