/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.location.elmo.impl;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.semantic.EntityFactory;
import it.tidalwave.semantic.IdFactory;
import it.tidalwave.semantic.IdentityWeakHashMap;
import it.tidalwave.geo.location.GeoLocation;
import it.tidalwave.geo.location.GeoSchema;
import it.tidalwave.geo.location.elmo.GeoVocabulary;
import it.tidalwave.netbeans.util.Locator;
import static it.tidalwave.role.Displayable.Displayable;
import static it.tidalwave.semantic.OwlThing.OwlThing;
import static it.tidalwave.semantic.SkosConcept.SkosConcept;
import static it.tidalwave.geo.location.elmo.impl.Support.doNothing;
import static it.tidalwave.geo.location.GeoLocation.GeoLocation;

/***********************************************************************************************************************
 *
 * @author Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@SuppressWarnings("PMD.TooManyStaticImports")
public class ElmoGeoLocationFactory 
  {
    private final static String CLASS = ElmoGeoLocationFactory.class.getName();
    private final static Logger logger = Logger.getLogger(CLASS);

    private static final EntityFactory<GeoLocation, ElmoGeoLocation> FACTORY =
                         EntityFactory.<GeoLocation, ElmoGeoLocation>
                                       forInterface(GeoLocation.class).
                                       withImplementation(ElmoGeoLocation.class).
                                       andRoles(OwlThing, SkosConcept).create();

    /** Used for performance. TODO: check if it's actually needed. */
    private final static IdentityWeakHashMap<GeoCoderEntityProxy, GeoLocation> CACHE =
                     new IdentityWeakHashMap<GeoCoderEntityProxy, GeoLocation>();

    /*******************************************************************************************************************
     *
     * Given a schema and an entity proxy, returns a location. If it already exists in the cache, is returned as is;
     * otherwise a semantically-equivalent location is searched and returned; otherwise, a new one is created.
     *
     * @param  geoSchema       the schema
     * @param  gcEntityProxy   the entity proxy
     * @return                 the location
     *
     ******************************************************************************************************************/
    @Nonnull
    public static synchronized GeoLocation findOrCreate (@CheckForNull final GeoSchema geoSchema,
                                                         @CheckForNull final GeoCoderEntityProxy gcEntityProxy)
      {
        logger.fine("findOrCreate(%s, %s)", geoSchema, gcEntityProxy);

        GeoLocation geoLocation = null;

        if (gcEntityProxy != null)
          {
            geoLocation = findSameAs(gcEntityProxy);
          }

        if (geoLocation == null)
          {
            final IdFactory idFactory = Locator.find(IdFactory.class);
            geoLocation = FACTORY.create(idFactory.createId(GeoVocabulary.ID_PREFIX_LOCATION));

            if (gcEntityProxy != null)
              {
                geoLocation.as(OwlThing).addSameAs(gcEntityProxy);
                geoLocation.as(SkosConcept).setDisplayName(gcEntityProxy.getGeoCoderEntity().as(Displayable).getDisplayName());
              }
          }

        if (gcEntityProxy != null) // FIXME: do only if it's a new thing
          {
            CACHE.put(gcEntityProxy, geoLocation);
          }

        return geoLocation;
      }

    /*******************************************************************************************************************
     *
     * Clears all flyweight caches (used for testing).
     *
     ******************************************************************************************************************/
    public static void flushCaches()
      {
        logger.info("flushCaches()");
        CACHE.clear();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @CheckForNull
    private static GeoLocation findSameAs (final @Nonnull GeoCoderEntityProxy gcEntityProxy)
      {
        GeoLocation geoLocation = CACHE.get(gcEntityProxy);

        if (geoLocation == null)
          {
            try
              {
                geoLocation = gcEntityProxy.as(OwlThing).findSameAs().ofType(GeoLocation).result();
                CACHE.put(gcEntityProxy, geoLocation);
              }
            catch (NotFoundException e)
              {
                doNothing(); // don't worry, go on
              }
          }

        return geoLocation;
      }
  }
