/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.location.elmo.impl;

import javax.annotation.Nonnull;
import java.util.Arrays;
import java.util.Locale;
import java.io.File;
import org.openide.util.Lookup;
import org.openrdf.elmo.Entity;
import org.openrdf.repository.sail.SailRepositoryConnection;
import org.openrdf.rio.rdfxml.util.OrganizedRDFXMLWriter;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.util.As;
import it.tidalwave.util.Id;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.semantic.Wrapper;
import it.tidalwave.geo.Coordinate;
import it.tidalwave.geo.geocoding.GeoCoderEntity.Type;
import it.tidalwave.geo.location.GeoLocation;
import it.tidalwave.geo.location.GeoPath;
import it.tidalwave.geo.location.GeoSchema;
import it.tidalwave.geo.location.GeoSchemaManager;
import it.tidalwave.geo.location.elmo.GeoVocabulary;
import it.tidalwave.role.Displayable;
import it.tidalwave.role.MutableDisplayable;
import static it.tidalwave.role.Displayable.Displayable;
import org.junit.Test;
import static org.junit.Assert.*;
import static org.hamcrest.CoreMatchers.*;

/***********************************************************************************************************************
 *
 * Tests the functions of the API starting from a scratch repository, populating it with data from the remote GeoCoder.
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class ElmoGeoLocationIntegrationTestSupport extends TestSupport
  {
    private final static String CLASS = ElmoGeoLocationIntegrationTestSupport.class.getName();
    private final static Logger logger = Logger.getLogger(CLASS);

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Test
    public void shouldProperlyRetrieveDataForItaly()
      throws Exception
      {
        final GeoSchemaManager schemaManager = Locator.find(GeoSchemaManager.class);
        final GeoSchema geoSchema = schemaManager.findSchemaByName(GEO_CODER_NAME);
        assertThat(geoSchema, is(notNullValue()));
        assertThat(Wrapper.isWrapped(geoSchema), is(true));
        assertThat(Wrapper.unwrap(geoSchema), is(notNullValue()));

        final GeoLocation earth = geoSchema.getRoot();
        assertGood(earth, GeoLocation.class);
        assertThat(earth.as(Displayable).getDisplayName(), is("Earth"));
        assertThat(earth.findPath(geoSchema).asList(), is(Arrays.asList(earth)));
        assertThat(earth.findCoordinate(geoSchema), is(new Coordinate(0, 0, 0)));
        assertThat(earth.findCode(geoSchema), is(""));
        assertThat(earth.findType(geoSchema), is((Object)Type.PLANET));
// FIXME        assertThat(earth, is(sameInstance(geoSchema.findLocations().qualifiedName("/").result())));

        final GeoLocation europe = earth.findChildren(geoSchema).withDisplayName("Europe").result();
        assertGood(europe, GeoLocation.class);
        assertThat(europe.as(Displayable).getDisplayName(), is("Europe"));
        assertThat(europe.findParent(geoSchema), is(earth));
        assertThat(europe.findPath(geoSchema).asList(), is(Arrays.asList(earth, europe)));
        assertThat(europe.findCoordinate(geoSchema), is(new Coordinate(48.69096, 9.14062)));
        assertThat(europe.findCode(geoSchema), is("EU"));
        assertThat(europe.findType(geoSchema), is((Object)Type.CONTINENT));
// FIXME        assertThat(europe, is(sameInstance(geoSchema.findLocations().qualifiedName("/Europe").result())));

        final GeoLocation italy = europe.findChildren(geoSchema).withDisplayName("Italy").result();
        assertGood(italy, GeoLocation.class);
        assertThat(italy.as(Displayable).getDisplayName(), is("Italy"));
        assertThat(italy.findParent(geoSchema), is(europe));
        assertThat(italy.findPath(geoSchema).asList(), is(Arrays.asList(earth, europe, italy)));
        assertThat(italy.findCoordinate(geoSchema), is(new Coordinate(42.83333, 12.83333, 0)));
        assertThat(italy.findCode(geoSchema), is("IT"));
        assertThat(italy.findType(geoSchema), is((Object)Type.COUNTRY));
// FIXME        assertThat(italy, is(sameInstance(geoSchema.findLocations().qualifiedName("/Europe/Italy").result())));

        final GeoLocation liguria = italy.findChildren(geoSchema).withDisplayName("Liguria").result();
        assertGood(liguria, GeoLocation.class);
        assertThat(liguria.as(Displayable).getDisplayName(), is("Liguria"));
        assertThat(liguria.findParent(geoSchema), is(italy));
        assertThat(liguria.findPath(geoSchema).asList(), is(Arrays.asList(earth, europe, italy, liguria)));
        assertThat(liguria.findCoordinate(geoSchema), is(new Coordinate(44.5, 8.83333, 0)));
        assertThat(liguria.findCode(geoSchema), is("08")); 
        assertThat(liguria.findType(geoSchema), is((Object)Type.ADMIN_DIVISION_1));
// FIXME        assertThat(liguria, is(sameInstance(geoSchema.findLocations().qualifiedName("/Europe/Italy/Liguria").result())));

        final GeoLocation provinciaDiGenova = liguria.findChildren(geoSchema).withDisplayName("Genoa").result();
        assertGood(provinciaDiGenova, GeoLocation.class);
        assertThat(provinciaDiGenova.as(Displayable).getDisplayName(), is("Genoa"));
        assertThat(provinciaDiGenova.findParent(geoSchema), is(liguria));
        assertThat(provinciaDiGenova.findPath(geoSchema).asList(), is(Arrays.asList(earth, europe, italy, liguria, provinciaDiGenova)));
        assertThat(provinciaDiGenova.findCoordinate(geoSchema), is(new Coordinate(44.5, 9.06667, 0)));
        assertThat(provinciaDiGenova.findCode(geoSchema), is("GE"));
        assertThat(provinciaDiGenova.findType(geoSchema), is((Object)Type.ADMIN_DIVISION_2));
// FIXME        assertThat(provinciaDiGenova, is(sameInstance(geoSchema.findLocations().qualifiedName("/Europe/Italy/Liguria/Genoa").result())));

        final GeoLocation recco = provinciaDiGenova.findChildren(geoSchema).withDisplayName("Recco").result();
        assertGood(recco, GeoLocation.class);
        assertThat(recco.as(Displayable).getDisplayName(), is("Recco"));
        assertThat(recco.findParent(geoSchema), is(provinciaDiGenova));
        assertThat(recco.findPath(geoSchema).asList(), is(Arrays.asList(earth, europe, italy, liguria, provinciaDiGenova, recco)));
        assertThat(recco.findCoordinate(geoSchema), is(new Coordinate(44.36324, 9.13717, 0)));
        assertThat(recco.findCode(geoSchema), is(""));
        assertThat(recco.findType(geoSchema), is((Object)Type.ADMIN_DIVISION_3));
// FIXME        assertThat(recco, is(sameInstance(geoSchema.findLocations().qualifiedName("/Europe/Italy/Liguria/Genoa/Recco").result())));

        final GeoPath path = recco.findPath(geoSchema);
        assertThat(path.findByType(geoSchema, Type.PLANET), is(earth));
        assertThat(path.findByType(geoSchema, Type.CONTINENT), is(europe));
        assertThat(path.findByType(geoSchema, Type.COUNTRY), is(italy));
        assertThat(path.findByType(geoSchema, Type.ADMIN_DIVISION_1), is(liguria));
        assertThat(path.findByType(geoSchema, Type.ADMIN_DIVISION_2), is(provinciaDiGenova));

        final Coordinate polanesiCoordinates = new Coordinate(44.36667, 9.11667);
        final GeoLocation polanesi = recco.createChild().
                                           withDisplayName(Locale.ITALIAN, "Polanesi").
                                           withDisplayName(Locale.FRENCH, "Polanesi").
                                           withDisplayName(Locale.ENGLISH, "Polanesi").
                                           withCode("XYZ").
                                           at(polanesiCoordinates).build();
        assertNotNull(polanesi);
        assertTrue(Wrapper.isWrapped(polanesi));
        assertThat(polanesi.as(Displayable).getDisplayName(), is("Polanesi"));
        assertThat(polanesi.findParent(geoSchema), is(recco));
        assertThat(polanesi.findPath(geoSchema).asList(), is(Arrays.asList(earth, europe, italy, liguria, provinciaDiGenova, recco, polanesi)));
        assertThat(polanesi.findCoordinate(geoSchema), is(polanesiCoordinates));
        assertThat(polanesi.findCode(geoSchema), is("XYZ"));
        assertThat(polanesi.findType(geoSchema), is((Object)Type.UNKNOWN));

        dump("it-pre"); // FIXME: compare with the expected dump so we also assert there are no unrequested children

//        assertEquals(0, recco.findChildren(schema).size()); FIXME screws up things
        assertThat(provinciaDiGenova.findChildren(geoSchema).count(), is(66));
        assertThat(liguria.findChildren(geoSchema).count(), is(4));
        assertThat(italy.findChildren(geoSchema).count(), is(20));
        assertThat(europe.findChildren(geoSchema).count(), is(50));
        assertThat(earth.findChildren(geoSchema).count(), is(7));

        dump("it-final");
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Test
    public void shouldProperlyRetrieveDataForFrance()
      throws Exception
      {
        final GeoSchemaManager schemaManager = Locator.find(GeoSchemaManager.class);
        final GeoSchema geoNamesSchema = schemaManager.findSchemaByName("GeoNames");
        assertNotNull(geoNamesSchema);
        assertTrue(Wrapper.isWrapped(geoNamesSchema));
        assertNotNull(Wrapper.unwrap(geoNamesSchema));

        final GeoLocation earth = geoNamesSchema.getRoot();
        assertGood(earth, GeoLocation.class);
        assertEquals("Earth", earth.as(Displayable).getDisplayName());
        assertEquals(Arrays.asList(earth), earth.findPath(geoNamesSchema).asList());

        final GeoLocation europe = earth.findChildren(geoNamesSchema).withDisplayName("Europe").result();
        assertGood(europe, GeoLocation.class);
        assertEquals("Europe", europe.as(Displayable).getDisplayName());
        assertEquals(earth, europe.findParent(geoNamesSchema));
        assertEquals(Arrays.asList(earth, europe), europe.findPath(geoNamesSchema).asList());

        final GeoLocation france = europe.findChildren(geoNamesSchema).withDisplayName("France").result();
        assertNotNull(france);
        assertTrue(Wrapper.isWrapped(france));
        assertEquals("France", france.as(Displayable).getDisplayName());
        assertEquals(europe, france.findParent(geoNamesSchema));
        assertEquals(Arrays.asList(earth, europe, france), france.findPath(geoNamesSchema).asList());
        assertEquals("FR", france.findCode(geoNamesSchema));
        assertEquals(Type.COUNTRY, france.findType(geoNamesSchema));

        final GeoLocation bourgogne = france.findChildren(geoNamesSchema).withDisplayName("Bourgogne").result();
        assertNotNull(bourgogne);
        assertTrue(Wrapper.isWrapped(bourgogne));
        assertEquals("Bourgogne", bourgogne.as(Displayable).getDisplayName());
        assertEquals(france, bourgogne.findParent(geoNamesSchema));
        assertEquals(Arrays.asList(earth, europe, france, bourgogne), bourgogne.findPath(geoNamesSchema).asList());
        assertEquals("A1", bourgogne.findCode(geoNamesSchema));
        assertEquals(Type.ADMIN_DIVISION_1, bourgogne.findType(geoNamesSchema));

        final GeoLocation yonne = bourgogne.findChildren(geoNamesSchema).withDisplayName("Yonne").result();
        assertNotNull(yonne);
        assertTrue(Wrapper.isWrapped(yonne));
        assertEquals("Yonne", yonne.as(Displayable).getDisplayName());
        assertEquals(bourgogne, yonne.findParent(geoNamesSchema));
        assertEquals(Arrays.asList(earth, europe, france, bourgogne, yonne), yonne.findPath(geoNamesSchema).asList());
        assertEquals("89", yonne.findCode(geoNamesSchema));
        assertEquals(Type.ADMIN_DIVISION_2, yonne.findType(geoNamesSchema));

        final GeoLocation arrAvallon = yonne.findChildren(geoNamesSchema).withDisplayName("Arrondissement d'Avallon").result();
        assertNotNull(arrAvallon);
        assertTrue(Wrapper.isWrapped(arrAvallon));
        assertEquals("Arrondissement d'Avallon", arrAvallon.as(Displayable).getDisplayName());
        assertEquals(yonne, arrAvallon.findParent(geoNamesSchema));
        assertEquals(Arrays.asList(earth, europe, france, bourgogne, yonne, arrAvallon), arrAvallon.findPath(geoNamesSchema).asList());
        assertEquals("", arrAvallon.findCode(geoNamesSchema));
        assertEquals(Type.ADMIN_DIVISION_3, arrAvallon.findType(geoNamesSchema));

        final GeoLocation vezelay = arrAvallon.findChildren(geoNamesSchema).withDisplayName("V\u00e9zelay").result();
        assertNotNull(vezelay);
        assertTrue(Wrapper.isWrapped(vezelay));
        assertEquals("V\u00e9zelay", vezelay.as(Displayable).getDisplayName());
        assertEquals(arrAvallon, vezelay.findParent(geoNamesSchema));
        assertEquals(Arrays.asList(earth, europe, france, bourgogne, yonne, arrAvallon, vezelay), vezelay.findPath(geoNamesSchema).asList());
        assertEquals("", vezelay.findCode(geoNamesSchema));
        assertEquals(Type.ADMIN_DIVISION_4, vezelay.findType(geoNamesSchema));

        final GeoLocation vezelay2 = vezelay.findChildren(geoNamesSchema).withDisplayName("V\u00e9zelay").result();
        assertNotNull(vezelay2);
        assertTrue(Wrapper.isWrapped(vezelay2));
        assertEquals("V\u00e9zelay", vezelay2.as(Displayable).getDisplayName());
        assertEquals(vezelay, vezelay2.findParent(geoNamesSchema));
        assertEquals(Arrays.asList(earth, europe, france, bourgogne, yonne, arrAvallon, vezelay, vezelay2), vezelay2.findPath(geoNamesSchema).asList());
        assertEquals("", vezelay2.findCode(geoNamesSchema));
        assertEquals(Type.INHABITED_PLACE, vezelay2.findType(geoNamesSchema));

        final GeoPath path = vezelay2.findPath(geoNamesSchema);
        assertEquals(earth,      path.findByType(geoNamesSchema, Type.PLANET));
        assertEquals(europe,     path.findByType(geoNamesSchema, Type.CONTINENT));
        assertEquals(france,     path.findByType(geoNamesSchema, Type.COUNTRY));
        assertEquals(bourgogne,  path.findByType(geoNamesSchema, Type.ADMIN_DIVISION_1));
        assertEquals(yonne,      path.findByType(geoNamesSchema, Type.ADMIN_DIVISION_2));
        assertEquals(arrAvallon, path.findByType(geoNamesSchema, Type.ADMIN_DIVISION_3));
        assertEquals(vezelay,    path.findByType(geoNamesSchema, Type.ADMIN_DIVISION_4));

        dump("fr-pre"); // FIXME: assert there are no unrequested children

        assertEquals(  0, vezelay2.findChildren(geoNamesSchema).count());
        assertEquals(  1, vezelay.findChildren(geoNamesSchema).count());
        assertEquals(149, arrAvallon.findChildren(geoNamesSchema).count());
        assertEquals(  3, yonne.findChildren(geoNamesSchema).count());
        assertEquals(  4, bourgogne.findChildren(geoNamesSchema).count());
        assertEquals( 26, france.findChildren(geoNamesSchema).count());
        assertEquals( 50, europe.findChildren(geoNamesSchema).count());
        assertEquals(  7, earth.findChildren(geoNamesSchema).count());

        dump("fr-final");
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Test
    public void shouldProperlyRetrieveDataForUSA()
      throws Exception
      {
        final GeoSchemaManager schemaManager = Locator.find(GeoSchemaManager.class);
        final GeoSchema geoNamesSchema = schemaManager.findSchemaByName("GeoNames");
        assertNotNull(geoNamesSchema);
        assertTrue(Wrapper.isWrapped(geoNamesSchema));
        assertNotNull(Wrapper.unwrap(geoNamesSchema));

        final GeoLocation earth = geoNamesSchema.getRoot();
        assertNotNull(earth);
        assertTrue(Wrapper.isWrapped(earth));
        assertEquals("Earth", earth.as(Displayable).getDisplayName());
        assertNotNull(Wrapper.unwrap(earth));
        assertEquals(Arrays.asList(earth), earth.findPath(geoNamesSchema).asList());

        final GeoLocation northAmerica = earth.findChildren(geoNamesSchema).withDisplayName("North America").result();
        assertNotNull(northAmerica);
        assertTrue(Wrapper.isWrapped(northAmerica));
        assertEquals("North America", northAmerica.as(Displayable).getDisplayName());
        assertEquals(earth, northAmerica.findParent(geoNamesSchema));
        assertEquals(Arrays.asList(earth, northAmerica), northAmerica.findPath(geoNamesSchema).asList());
        assertEquals("NA", northAmerica.findCode(geoNamesSchema));
        assertEquals(Type.CONTINENT, northAmerica.findType(geoNamesSchema));

        final GeoLocation usa = northAmerica.findChildren(geoNamesSchema).withDisplayName("United States").result();
        assertNotNull(usa);
        assertTrue(Wrapper.isWrapped(usa));
        assertEquals("United States", usa.as(Displayable).getDisplayName());
        assertEquals(northAmerica, usa.findParent(geoNamesSchema));
        assertEquals(Arrays.asList(earth, northAmerica, usa), usa.findPath(geoNamesSchema).asList());
        assertEquals("US", usa.findCode(geoNamesSchema));
        assertEquals(Type.COUNTRY, usa.findType(geoNamesSchema));

        final GeoLocation california = usa.findChildren(geoNamesSchema).withDisplayName("California").result();
        assertNotNull(california);
        assertTrue(Wrapper.isWrapped(california));
        assertEquals("California", california.as(Displayable).getDisplayName());
        assertEquals(usa, california.findParent(geoNamesSchema));
        assertEquals(Arrays.asList(earth, northAmerica, usa, california), california.findPath(geoNamesSchema).asList());
        assertEquals("CA", california.findCode(geoNamesSchema));
        assertEquals(Type.ADMIN_DIVISION_1, california.findType(geoNamesSchema));

        final GeoLocation marinCounty = california.findChildren(geoNamesSchema).withDisplayName("Marin County").result();
        assertNotNull(marinCounty);
        assertTrue(Wrapper.isWrapped(marinCounty));
        assertEquals("Marin County", marinCounty.as(Displayable).getDisplayName());
        assertEquals(california, marinCounty.findParent(geoNamesSchema));
        assertEquals(Arrays.asList(earth, northAmerica, usa, california, marinCounty), marinCounty.findPath(geoNamesSchema).asList());
        assertEquals("041", marinCounty.findCode(geoNamesSchema));
        assertEquals(Type.ADMIN_DIVISION_2, marinCounty.findType(geoNamesSchema));

        final GeoLocation sausalito = marinCounty.findChildren(geoNamesSchema).withDisplayName("Sausalito").result();
        assertNotNull(sausalito);
        assertTrue(Wrapper.isWrapped(sausalito));
        assertEquals("Sausalito", sausalito.as(Displayable).getDisplayName());
        assertEquals(marinCounty, sausalito.findParent(geoNamesSchema));
        assertEquals(Arrays.asList(earth, northAmerica, usa, california, marinCounty, sausalito), sausalito.findPath(geoNamesSchema).asList());
        assertEquals("", sausalito.findCode(geoNamesSchema));
        assertEquals(Type.INHABITED_PLACE, sausalito.findType(geoNamesSchema));

        final GeoPath path = sausalito.findPath(geoNamesSchema);
        assertEquals(earth, path.findByType(geoNamesSchema, Type.PLANET));
        assertEquals(northAmerica, path.findByType(geoNamesSchema, Type.CONTINENT));
        assertEquals(usa, path.findByType(geoNamesSchema, Type.COUNTRY));
        assertEquals(california, path.findByType(geoNamesSchema, Type.ADMIN_DIVISION_1));
        assertEquals(marinCounty, path.findByType(geoNamesSchema, Type.ADMIN_DIVISION_2));

        dump("us-pre"); // FIXME: assert there are no unrequested children

        assertEquals(0,  sausalito.findChildren(geoNamesSchema).count());
        assertEquals(115, marinCounty.findChildren(geoNamesSchema).count());
        assertEquals(58, california.findChildren(geoNamesSchema).count());
        assertEquals(51, usa.findChildren(geoNamesSchema).count());
        assertEquals(41, northAmerica.findChildren(geoNamesSchema).count());
        assertEquals(7,  earth.findChildren(geoNamesSchema).count());

        dump("us-final");
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void dump (@Nonnull final String name)
      throws Exception
      {
        final File targetFolder = new File("target/test-artifacts");
        final File srcFolder = new File("src/test/resources");
        targetFolder.mkdirs();
        
        final String fileName = "Geo-" + name + ".rdf";
        final File dumpedFile = new File(targetFolder, fileName);
        final File expectedFile = new File(srcFolder, fileName);

        final OrganizedRDFXMLWriter writer = new OrganizedRDFXMLWriter(dumpedFile, "UTF-8");
        writer.handleNamespace("dcterms", "http://purl.org/dc/terms/");
        writer.handleNamespace("geo", GeoVocabulary.NS_GEO);
        writer.handleNamespace("wgs84_pos", GeoVocabulary.NS_WGS84_POS);
        writer.handleNamespace("skos", "http://www.w3.org/2004/02/skos/core#");
        writer.handleNamespace("owl", "http://www.w3.org/2002/07/owl#");

        final SailRepositoryConnection connection = sailRepository.getConnection();
        connection.export(writer);
        connection.close();

//        AssertFileContents.assertSameFiles(expectedFile, dumpedFile); FIXME
      }

    /*******************************************************************************************************************
     *
     * FIXME: make it a matcher
     *
     ******************************************************************************************************************/
    public static <T extends Lookup.Provider> void assertGood (@Nonnull final T object,
                                                               @Nonnull final Class<? extends T> interfaceClass)
      {
        assertNotNull(object);
        assertTrue(Wrapper.isWrapped(object));
        final T unwrapped = Wrapper.unwrap(object);
        assertNotNull(unwrapped);
        assertTrue(object instanceof Entity);
        assertTrue(interfaceClass.isAssignableFrom(object.getClass()));
        assertTrue(interfaceClass.isAssignableFrom(unwrapped.getClass()));
        final Lookup lookup = object.getLookup();
        assertNotNull(lookup);
        final Entity entity = lookup.lookup(Entity.class);
        assertNotNull(entity);
        final Id id = lookup.lookup(Id.class);
        assertNotNull(id);
        final As as = lookup.lookup(As.class);
        assertNotNull(as);
        final Displayable displayable = lookup.lookup(Displayable.class);
        assertNotNull(displayable);
        final MutableDisplayable mutableDisplayable = lookup.lookup(MutableDisplayable.class);
        assertNotNull(mutableDisplayable);
      }
  }
