/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.location.elmo.impl;

import javax.annotation.Nonnull;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import org.mockito.invocation.InvocationOnMock;
import org.openrdf.repository.sail.SailRepositoryConnection;
import org.openrdf.rio.RDFFormat;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.geo.Coordinate;
import it.tidalwave.geo.geocoding.GeoCoder;
import it.tidalwave.geo.geocoding.GeoCoderEntity.Type;
import it.tidalwave.geo.geocoding.GeoCoderManager;
import it.tidalwave.netbeans.util.Locator;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.AfterClass;
import junit.framework.AssertionFailedError;
import it.tidalwave.netbeans.util.test.MockLookup;
import org.mockito.stubbing.Answer;
import static org.mockito.Mockito.*;
import static org.junit.Assert.assertThat;
import static org.hamcrest.Matchers.is;
import static org.hamcrest.Matchers.sameInstance;

/***********************************************************************************************************************
 *
 * Tests the functions of the API starting from a populated repository, verifying that the remote GeoCoder is not
 * called.
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class ElmoGeoLocationOfflineIntegrationTest extends ElmoGeoLocationIntegrationTestSupport
  {
    private final static String CLASS = ElmoGeoLocationOfflineIntegrationTest.class.getName();
    private final static Logger logger = Logger.getLogger(CLASS);

    @BeforeClass
    public static void installGeoCoderPreventer()
      throws NotFoundException
      {
        logger.info("********** SETTING UP A GEOCODER PREVENTER");

        final GeoCoderManager geoCoderManager = Locator.find(GeoCoderManager.class);
        final GeoCoder realGeoCoder = geoCoderManager.findGeoCoderByName(GEO_CODER_NAME);
        assertThat(realGeoCoder.getClass().getName(), is("it.tidalwave.geo.geocoding.geonamesprovider.GeoNamesProvider"));
        
        final GeoCoder geoCoderPreventer = mock(GeoCoder.class);
        final Answer<Object> shouldNotGetHere = new Answer<Object>()
          {
            @Override
            public Object answer (final @Nonnull InvocationOnMock invocation)
              {                 
                throw new AssertionFailedError("Should not get here: " + invocation);
              }
          };

        when(geoCoderPreventer.getName()).thenReturn(realGeoCoder.getName());
        when(geoCoderPreventer.getRoot()).thenReturn(realGeoCoder.getRoot());
        // FIXME: try to delegate to the real geoCoder
        when(geoCoderPreventer.typeFromString(eq("PLANET"))).thenReturn(Type.PLANET);
        when(geoCoderPreventer.typeFromString(eq("CONT"))).thenReturn(Type.CONTINENT);
        when(geoCoderPreventer.typeFromString(eq("PCLI"))).thenReturn(Type.COUNTRY);
        when(geoCoderPreventer.typeFromString(eq("ADM1"))).thenReturn(Type.ADMIN_DIVISION_1);
        when(geoCoderPreventer.typeFromString(eq("ADM2"))).thenReturn(Type.ADMIN_DIVISION_2);
        when(geoCoderPreventer.typeFromString(eq("ADM3"))).thenReturn(Type.ADMIN_DIVISION_3);
        when(geoCoderPreventer.typeFromString(eq("ADM4"))).thenReturn(Type.ADMIN_DIVISION_4);
        when(geoCoderPreventer.typeFromString(eq("PPL"))).thenReturn(Type.INHABITED_PLACE);
        // END FIXME

        when(geoCoderPreventer.findChildren(anyString())).thenAnswer(shouldNotGetHere);
        when(geoCoderPreventer.findCountryCode(any(Coordinate.class))).thenAnswer(shouldNotGetHere);
        when(geoCoderPreventer.findGeoEntityById(anyString())).thenAnswer(shouldNotGetHere);
        when(geoCoderPreventer.findEntities(any(Coordinate.class), anyFloat())).thenAnswer(shouldNotGetHere);
        when(geoCoderPreventer.findNearbyPostalCodes(any(Coordinate.class), anyFloat())).thenAnswer(shouldNotGetHere);
        when(geoCoderPreventer.findNeighbours(anyString())).thenAnswer(shouldNotGetHere);
        when(geoCoderPreventer.findSiblings(anyString())).thenAnswer(shouldNotGetHere);
//        when(geoCoderPreventer.findSubdivision(any(Coordinate.class))).thenAnswer(shouldNotGetHere);
        
        MockLookup.setInstances(geoCoderPreventer, TestIdFactory.getInstance());
        MockLookup.addBannedObject(realGeoCoder);

        assertThat(geoCoderManager.findGeoCoderByName(GEO_CODER_NAME), sameInstance(geoCoderPreventer));
      }

    @AfterClass
    public static void restoreGeoCoder()
      {
        MockLookup.reset();
      }

    @Before
    public void importDataFromFiles()
      throws Exception
      {
        logger.info("********** IMPORTING DATA INTO THE REPOSITORY");
        final SailRepositoryConnection connection = sailRepository.getConnection();
        final Reader r = getReader("src/test/resources/Geo-all-final.rdf");
        connection.add(r, "urn:foobar", RDFFormat.RDFXML);
        r.close();
        connection.close();
      }

    @Nonnull
    private static Reader getReader (final @Nonnull String name)
       throws IOException
      {
        return new InputStreamReader(new FileInputStream(name), "UTF-8");
      }
  }
