/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.location.node.impl;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import java.util.List;
import javax.swing.Action;
import java.awt.Image;
import org.openide.util.Parameters;
import org.openide.util.lookup.Lookups;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import it.tidalwave.geo.location.GeoLocation;
import it.tidalwave.geo.location.GeoSchema;
import it.tidalwave.netbeans.nodes.SafeChildFactory;
import it.tidalwave.role.Displayable;

/***********************************************************************************************************************
 *
 * A {@link Node} specialization for representing a {@link GeoEntity}.
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@edu.umd.cs.findbugs.annotations.SuppressWarnings("HE_INHERITS_EQUALS_USE_HASHCODE, EQ_DOESNT_OVERRIDE_EQUALS")
public final class GeoLocationNode extends AbstractNode
  {
    @Nonnull
    private final GeoLocationNodeFactoryImpl nodeFactory;

    @Nonnull
    private final GeoLocation geoLocation;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static class GeoLocationChildFactory extends SafeChildFactory<GeoLocation>
      {
        @Nonnull
        private final GeoSchema geoSchema;

        @Nonnull
        private final GeoLocation geoLocation;
        
        @Nonnull
        private final GeoLocationNodeFactoryImpl nodeFactory;

        public GeoLocationChildFactory (final @Nonnull GeoSchema geoSchema,
                                        final @Nonnull GeoLocation geoLocation,
                                        final @Nonnull GeoLocationNodeFactoryImpl nodeFactory)
          {
            assert geoSchema != null   : "geoSchema is null";
            assert geoLocation != null : "geoLocation is null";
            assert nodeFactory != null : "nodeFactory is null";

            this.geoSchema   = geoSchema;
            this.geoLocation = geoLocation;
            this.nodeFactory = nodeFactory;
          }

        @Override
        protected boolean createKeysSafely (final @Nonnull List<GeoLocation> geoLocations)
          {
            geoLocations.addAll(geoLocation.findChildren(geoSchema).results());
            return true;
          }
        
        @Override
        @Nonnull 
        protected Node createNodeForKey (final @Nonnull GeoLocation geoLocation)
          {
            return new GeoLocationNode(geoSchema, geoLocation, nodeFactory);
          }
        
        public void refresh()
          {
            refresh(false);
          }
      };
            
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public GeoLocationNode (final @Nonnull GeoSchema geoSchema,
                            final @Nonnull GeoLocation geoLocation,
                            final @Nonnull GeoLocationNodeFactoryImpl nodeFactory)
      {
        this(geoLocation, nodeFactory, new GeoLocationChildFactory(geoSchema, geoLocation, nodeFactory));
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private GeoLocationNode (final @Nonnull GeoLocation geoLocation,
                             final @Nonnull GeoLocationNodeFactoryImpl nodeFactory,
                             final @Nonnull GeoLocationChildFactory childFactory)
      {
        super(Children.create(childFactory, true), Lookups.singleton(geoLocation));
        this.nodeFactory = nodeFactory;
        this.geoLocation = geoLocation;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    @Nonnull
    public void setDisplayName (final @Nonnull String displayName)
      {
        Parameters.notNull("displayName", displayName);
        // TODO set it
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    @Nonnull
    public String getDisplayName()
      {
        return geoLocation.as(Displayable.class).getDisplayName();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    @Nonnull
    public String getName()
      {
        return geoLocation.as(Displayable.class).getDisplayName();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    @Nonnull
    public Image getIcon (final int type) 
      {
        return nodeFactory.findIcon(this);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    @Nonnull 
    public Image getOpenedIcon (final int type) 
      {
        return nodeFactory.findIcon(this);
      }   
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    @Nonnull 
    @edu.umd.cs.findbugs.annotations.SuppressWarnings("ITA_INEFFICIENT_TO_ARRAY")
    public Action[] getActions (final boolean context)
      {
        return nodeFactory.findNodeActions(getClass()).toArray(new Action[0]);
      }
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    @CheckForNull
    public Action getPreferredAction()
      {
        return nodeFactory.getPreferredNodeAction(getClass());
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public boolean canCopy()
      {
        return false;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public boolean canCut()
      {
        return false;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public boolean canDestroy()
      {
        return true;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public boolean canRename()
      {
        return true;
      }
  }
