/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.location.node.impl;

import java.beans.BeanInfo;
import java.util.Arrays;
import java.util.List;
import java.awt.Image;
import javax.swing.Action;
import org.mockito.invocation.InvocationOnMock;
import org.openide.nodes.Children;
import org.openide.util.Lookup;
import it.tidalwave.geo.location.GeoLocation;
import it.tidalwave.role.Displayable;
import org.hamcrest.CoreMatchers;
import org.junit.Before;
import org.junit.Test;
import org.mockito.stubbing.Answer;
import org.openide.nodes.Node;
import static org.junit.Assert.*;
import static org.hamcrest.CoreMatchers.*;
import static org.mockito.Mockito.*;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class GeoLocationNodeTest extends TestSupport
  {
    private static GeoLocationNodeFactoryImpl nodeFactory;
    private GeoLocationNode node;
    private GeoLocation parentGeoLocation;
    private GeoLocation childGeoLocation1;
    private GeoLocation childGeoLocation2;
    private Image icon;
    private List<? extends Action> actions;
    private Action preferredAction;

    @Before
    public void setUp()
      throws Exception
      {
        actions = Arrays.asList(mock(Action.class), mock(Action.class), mock(Action.class));
        parentGeoLocation = mock(GeoLocation.class);
        icon = mock(Image.class);
        preferredAction = mock(Action.class);
        final Displayable displayable = mock(Displayable.class);
        when(displayable.getDisplayName()).thenReturn("Italy");
        when(parentGeoLocation.as(eq(Displayable.class))).thenReturn(displayable);

        nodeFactory = mock(GeoLocationNodeFactoryImpl.class);
        doReturn(icon).when(nodeFactory).findIcon(argThat(CoreMatchers.any(Node.class)));
        doReturn(actions).when(nodeFactory).findNodeActions(eq(GeoLocationNode.class));
        doReturn(preferredAction).when(nodeFactory).getPreferredNodeAction(eq(GeoLocationNode.class));

        node = new GeoLocationNode(geoSchema, parentGeoLocation, nodeFactory);

        childGeoLocation1 = mock(GeoLocation.class);
        childGeoLocation2 = mock(GeoLocation.class);
        
        final GeoLocation.Finder finder = mock(GeoLocation.Finder.class);
//        when(finder.results()).thenReturn(Arrays.asList(c1, c2)); FIXME
        when(finder.results()).thenAnswer(new Answer<Object>() 
          {
            @Override
            public Object answer (final InvocationOnMock invocation)
              {
                return Arrays.asList(childGeoLocation1, childGeoLocation2);
              }
          });
        when(parentGeoLocation.findChildren(eq(geoSchema))).thenReturn(finder);
      }

    @Test
    public void shouldReturnChildrenNodesThatWrapsChildrenGeoLocations()
      throws InterruptedException
      {
        final Children children = node.getChildren();
        children.getNodes(); // triggers loading
        Thread.sleep(1000);  // waits for the "please wait..." node to disappear
        final Node[] nodes = children.getNodes();

        assertThat(nodes, is(notNullValue()));
        assertThat(nodes.length, is(2));
        assertThat(nodes[0].getLookup().lookup(GeoLocation.class), is(sameInstance(childGeoLocation1)));
        assertThat(nodes[1].getLookup().lookup(GeoLocation.class), is(sameInstance(childGeoLocation2)));
      }

    @Test
    public void shouldHaveGeoLocationInLookup()
      {
        final Lookup lookup = node.getLookup();
        assertThat(lookup, is(not(nullValue())));
        final GeoLocation lookedUpGeoEntity = lookup.lookup(GeoLocation.class);
        assertThat(lookedUpGeoEntity, sameInstance(parentGeoLocation));
      }

    @Test
    public void shouldBeDestroyableAndRenamableButNotCopiableOrCuttable()
      {
        assertThat("canCopy",    node.canCopy(),    is(false));
        assertThat("canCut",     node.canCut(),     is(false));
        assertThat("canDestroy", node.canDestroy(), is(true));
        assertThat("canRename",  node.canRename(),  is(true));
      }

    @Test
    public void shouldHaveTheSameNameAsTheGeoLocation()
      {
        final String name = node.getName();
        assertThat(name, equalTo("Italy"));
      }

    @Test
    public void shouldHaveTheSameDisplayNameAsTheGeoLocation()
      {
        final String displayName = node.getDisplayName();
        assertThat(displayName, equalTo("Italy"));
      }

    @Test
    public void shouldHaveANullHTMLDisplayName()
      {
        final String htmlDisplayName = node.getHtmlDisplayName();
        assertThat(htmlDisplayName, is(nullValue()));
      }

    @Test
    public void shouldReturnTheDefauktIcon()
      {
        final Image gotIcon = node.getIcon(BeanInfo.ICON_COLOR_16x16);
        assertThat(gotIcon, sameInstance(icon));
      }

    @Test
    public void shouldReturnTheDefaultOpenedIcon()
      {
        final Image gotIcon = node.getOpenedIcon(BeanInfo.ICON_COLOR_16x16);
        assertThat(gotIcon, sameInstance(icon));
      }

    @Test
    public void shouldReturnTheActions()
      {
        final Action[] gotActions = node.getActions(true);
        assertThat(gotActions, equalTo(actions.toArray(new Action[0])));
      }

    @Test
    public void shouldReturnThePreferredAction()
      {
        final Action gotAction = node.getPreferredAction();
        assertThat(gotAction, sameInstance(preferredAction));
      }
  }