/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.geocoding.geonamesprovider;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Level;
import java.io.Serializable;
import java.io.StringWriter;
import java.net.MalformedURLException;
import java.net.URL;
import javax.swing.Icon;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.geo.geocoding.GeoCoderEntity;
import it.tidalwave.geo.geocoding.GeoCoderEntity.Type;
import java.awt.Image;
import org.openide.util.ImageUtilities;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
class EmblemFactory implements GeoCoderEntity.EmblemFactory, Serializable
  {
    private static final String CLASS = EmblemFactory.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);
    private static final long serialVersionUID = 23459593059093534L;

    @Nonnull
    private final String id;

    @CheckForNull
    private transient Image emblem;

    private final AtomicBoolean loaded = new AtomicBoolean(false);

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public EmblemFactory (final @Nonnull String id)
      {
        this.id = id;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Image findEmblem (final @Nonnegative int size)
      throws NotFoundException
      {
        if (!loaded.getAndSet(true))
          {
            try
              {
                final Document document = GeoNames.retrieveDocument("hierarchy", "geonameId", id);
                final NodeList nodes = (NodeList)GeoNamesProvider.GEONAME_SET.evaluate(document, XPathConstants.NODESET);
                final Node node = nodes.item(nodes.getLength() - 1);
                log(document, node);
                final String typeAsString = (String)GeoNamesProvider.FCODE.evaluate(node, XPathConstants.STRING);
                final Type type = GeoNamesProvider.TYPE_MAP.get(typeAsString);
                URL url = null;

                if (type == Type.COUNTRY)
                  {
                    final String countryCode = (String)GeoNamesProvider.COUNTRY_CODE.evaluate(node, XPathConstants.STRING);
                    url = new URL(String.format("http://geotree.geonames.org/img/flags18/%s.png", countryCode));
                  }
                else if ((type != Type.PLANET) && (type != Type.CONTINENT))
                  {
                    final String countryCode = (String)GeoNamesProvider.COUNTRY_CODE.evaluate(node, XPathConstants.STRING);
                    url = new URL(String.format("http://geotree.geonames.org/img/blasons/%s/%s-100.png", countryCode, id));
                  }

                if (url == null)
                  {
                    throw new NotFoundException("No emblem for " + id);
                  }

                logger.finer(">>>> trying from %s...", url);
                Icon icon = GeoNames.retrieveIcon(url);
                logger.finer(">>>>>>>> found: %s", (icon == null) ? "no" : "yes");

                if (icon == null)
                  {
                    url = new URL(String.format("http://geotree.geonames.org/img/fcode18/%s.gif", typeAsString));
                    logger.finer(">>>> trying again from %s...", url);
                    icon = GeoNames.retrieveIcon(url);
                    logger.finer(">>>>>>>> found: %s", (icon == null) ? "no" : "yes");
                  }

                if (icon == null)
                  {
                    throw new NotFoundException("No emblem for " + id + " at " + url);
                  }

                emblem = ImageUtilities.icon2Image(icon);
              }
            catch (XPathExpressionException e)
              {
                throw new RuntimeException(e);
              }
            catch (MalformedURLException e)
              {
                throw new RuntimeException(e);
              }
          }

        if (emblem == null) // can be e.g. when loaded = true
          {
            throw new NotFoundException("No emblem for " + id);
          }
        
        return emblem;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void log (final @Nonnull Document document, final @Nonnull Node node)
      {
        if (logger.isLoggable(Level.FINEST))
          {
            try
              {
                final Transformer transformer = TransformerFactory.newInstance().newTransformer();
                final StringWriter sw = new StringWriter();
                transformer.transform(new DOMSource(document), new StreamResult(sw));
                logger.finest(">>>> document: %s", sw.getBuffer().toString());
                transformer.transform(new DOMSource(node), new StreamResult(sw));
                logger.finest(">>>> node: %s", sw.getBuffer().toString());
              }
            catch (TransformerException e)
              {
                e.printStackTrace();
              }
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public String toString()
      {
        return String.format("GeoNamesProvider$EmblemFactory[%s]", id);
      }
  }

