/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.geocoding.geonamesprovider;

import java.util.Arrays;
import java.util.SortedSet;
import java.util.TimeZone;
import java.util.TreeSet;
import java.util.Collection;
import java.util.Iterator;
import javax.xml.namespace.QName;
import java.awt.Image;
import org.openide.util.Lookup;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.geo.Coordinate;
import it.tidalwave.geo.geocoding.GeoCoder;
import it.tidalwave.geo.geocoding.GeoCoderEntity;
import it.tidalwave.geo.geocoding.GeoCoderEntity.Type;
import it.tidalwave.geo.geocoding.GeoCoderEntity.FactSheet;
import it.tidalwave.geo.geocoding.GeoCoderEntity.EmblemFactory;
import static it.tidalwave.role.Displayable.Displayable;
import static it.tidalwave.geo.geocoding.GeoCoderEntity.FactSheet.*;
import org.hamcrest.Matcher;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;
import static org.hamcrest.CoreMatchers.*;
import static org.junit.matchers.JUnitMatchers.*;
import it.tidalwave.netbeans.util.test.TestLoggerSetup;

/*******************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 ******************************************************************************/
public class GeoNamesProviderTest 
  {
    private static final String ID_EUROPE = "6255148";
    private GeoNamesProvider geoNamesProvider;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @BeforeClass
    public static void setupLogging()
      {
        TestLoggerSetup.setupLogging(TestLoggerSetup.class);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Before
    public void setupFixture()
      throws Exception 
      {
        geoNamesProvider = new GeoNamesProvider();
        GeoNames.clearCache();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Test
    public void mustBeRegisteredInMETA_INF()
      {
        final Matcher m = instanceOf(GeoNamesProvider.class);
        assertThat(Lookup.getDefault().lookupAll(GeoCoder.class), hasItem(m));
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Test
    public void getRoot()
      throws NotFoundException
      {
        log("getRoot");
        final GeoCoderEntity earth = geoNamesProvider.getRoot();
        assertNotNull(earth);
        assertEquals("Earth", earth.as(Displayable).getDisplayName());
//        assertEquals("Earth", earth.getLocalName());
        final Collection<? extends GeoCoderEntity> children = earth.findChildren().results();
        assertNotNull(children);
        assertEquals(7, children.size());
        final Iterator<? extends GeoCoderEntity> i = children.iterator();
        assertGeoEntity(i.next(), "6255146", "AF", Type.CONTINENT, "",   7.19, 21.09, "Africa");
        assertGeoEntity(i.next(), "6255152", "AN", Type.CONTINENT, "", -83.36, 16.52, "Antarctica");
        assertGeoEntity(i.next(), "6255147", "AS", Type.CONTINENT, "",  29.84, 89.30, "Asia");
        assertGeoEntity(i.next(), ID_EUROPE, "EU", Type.CONTINENT, "",  48.69,  9.14, "Europe");
        assertGeoEntity(i.next(), "6255149", "NA", Type.CONTINENT, "",  46.07,-100.55,"North America");
        assertGeoEntity(i.next(), "6255151", "OC", Type.CONTINENT, "", -18.31,138.52, "Oceania");
        assertGeoEntity(i.next(), "6255150", "SA", Type.CONTINENT, "", -14.60,-57.66, "South America");
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Test
    public void findChildren()
      throws NotFoundException
      {
        log("findChildren");
        final Collection<? extends GeoCoderEntity> result = geoNamesProvider.findChildren(ID_EUROPE).results();
        assertEquals(50, result.size());
        final Iterator<? extends GeoCoderEntity> i = result.iterator();
        assertGeoEntity(i.next(),  "783754", "AL", Type.COUNTRY, "?",  41.00, 20.00, "Albania");
        assertGeoEntity(i.next(), "3041565", "AD", Type.COUNTRY, "?",  42.50,  1.50, "Andorra");
        assertGeoEntity(i.next(), "2782113", "AT", Type.COUNTRY, "?",  47.33, 13.33, "Austria");
        assertGeoEntity(i.next(),  "630336", "BY", Type.COUNTRY, "?",  53.00, 28.00, "Belarus");
        assertGeoEntity(i.next(), "2802361", "BE", Type.COUNTRY, "?",  50.83,  4.00, "Belgium");
        assertGeoEntity(i.next(), "3277605", "BA", Type.COUNTRY, "?",  44.25, 17.83, "Bosnia and Herzegovina");
        
        for (int j = 0; j < 40; j++) // TODO too lazy to test all...
          {
            i.next();  
          }
        
        assertGeoEntity(i.next(),  "690791", "UA", Type.COUNTRY, "?",  49.00, 32.00, "Ukraine");
        assertGeoEntity(i.next(), "2635167", "GB", Type.COUNTRY, "?",  54.00, -4.00, "United Kingdom");
        assertGeoEntity(i.next(), "3164670", "VA", Type.COUNTRY, "?",  41.90225, 12.4533, "Vatican City");
        assertGeoEntity(i.next(),  "661882",   "", Type.UNKNOWN, "?",  60.25, 20.00, "Åland");
        assertThat(i.hasNext(), is(false));
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Test
    public void findGeoEntityById()
      throws NotFoundException
      {
        log("findGeoEntityById");
        GeoCoderEntity geoEntity = geoNamesProvider.findGeoEntityById("3179418");
        assertGeoEntity(geoEntity, "3179418", "",   Type.INHABITED_PLACE,  "?", 42.76, 10.88, "Castiglione della Pescaia");
        geoEntity = geoEntity.getParent();
        assertGeoEntity(geoEntity, "6536858", "",   Type.ADMIN_DIVISION_3, "?", 42.76, 10.88, "Castiglione della Pescaia");
        geoEntity = geoEntity.getParent();
        assertGeoEntity(geoEntity, "3175784", "GR", Type.ADMIN_DIVISION_2, "?", 42.83, 11.25, "Provincia di Grosseto");
        geoEntity = geoEntity.getParent();
        assertGeoEntity(geoEntity, "3165361", "16", Type.ADMIN_DIVISION_1, "?", 43.42, 11.00, "Tuscany");
        geoEntity = geoEntity.getParent();
        assertGeoEntity(geoEntity, "3175395", "IT", Type.COUNTRY,          "?", 42.83, 12.83, "Italy");
        geoEntity = geoEntity.getParent();
        assertGeoEntity(geoEntity, ID_EUROPE, "EU", Type.CONTINENT,        "",  48.69,  9.14, "Europe");
        geoEntity = geoEntity.getParent();
// FIXME        assertNull(geoEntity);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
////    @Test
//    public void findNeighbours()
//      {
//        log("findNeighbours");
//        Collection result = geoNamesProvider.findNeighbours(null);
//        fail("The test case is a prototype.");
//      }
//
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
//    @Test
    public void findSiblings()
      throws NotFoundException
      {
        log("findSiblings");
        Collection<? extends GeoCoderEntity> result = geoNamesProvider.findSiblings(ID_EUROPE).results(); // Europe
        assertEquals(5, result.size());
        final Iterator<? extends GeoCoderEntity> i = result.iterator();
        assertGeoEntity(i.next(), "3180116", "EU", Type.CONTINENT, "", 42.78, 10.83, "Europe");
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
////    @Test
//    public void findNearbyPostalCodes()
//      {
//        log("findNearbyPostalCodes");
//        Coordinate coordinates = new Coordinate(45, 11);
//        Collection result = geoNamesProvider.findNearbyPostalCodes(coordinates, 10);
//        fail("The test case is a prototype.");
//      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Test
    public void findNearestPlace()
      throws NotFoundException
      {
        log("findNearestPlace");
        Coordinate coordinates = new Coordinate(42.77, 10.84);
        final GeoCoderEntity gcEntity = geoNamesProvider.findNearestEntity(coordinates).result();
        assertGeoEntity(gcEntity, "3180116", "", Type.INHABITED_PLACE, "?", 42.79, 10.82, "Casa Mora", "Provincia di Grosseto", "Tuscany", "Italy");
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Test
    public void findNearbyPlaceNamesInSameProvince()
      throws NotFoundException
      {
        log("findNearbyPlaceNamesInSameProvince");
        Coordinate coordinates = new Coordinate(42.77, 10.84);
        Collection<? extends GeoCoderEntity> result = geoNamesProvider.findEntities(coordinates, 10).results();
        assertEquals(7, result.size());
        final Iterator<? extends GeoCoderEntity> i = result.iterator();
        assertGeoEntity(i.next(), "3180116", "", Type.INHABITED_PLACE, "?", 42.79, 10.82, "Casa Mora",                 "Provincia di Grosseto", "Tuscany", "Italy");
        assertGeoEntity(i.next(), "3179418", "", Type.INHABITED_PLACE, "?", 42.76, 10.88, "Castiglione della Pescaia", "Castiglione della Pescaia", "Provincia di Grosseto", "Tuscany", "Italy");
        assertGeoEntity(i.next(), "6544027", "", Type.INHABITED_PLACE, "?", 42.80, 10.76, "Punta Ala",                 "Provincia di Grosseto", "Tuscany", "Italy");
        assertGeoEntity(i.next(), "3171011", "", Type.INHABITED_PLACE, "?", 42.84, 10.82, "Pian d'Alma",               "Provincia di Grosseto", "Tuscany", "Italy");
        assertGeoEntity(i.next(), "3177162", "", Type.INHABITED_PLACE, "?", 42.78, 10.95, "Fattoria Badiola",          "Provincia di Grosseto", "Tuscany", "Italy");
        assertGeoEntity(i.next(), "3180370", "", Type.INHABITED_PLACE, "?", 42.85, 10.83, "Casa Agresti",              "Provincia di Grosseto", "Tuscany", "Italy");
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Test
    public void findNearbyPlaceNamesInDifferentCountries()
      throws NotFoundException
      {
        log("findNearbyPlaceNamesInDifferentCountries");
        Coordinate coordinates = new Coordinate(43.79, 7.52);
        Collection<? extends GeoCoderEntity> result = geoNamesProvider.findEntities(coordinates, 2).results();
        assertEquals(6, result.size());
        final Iterator<? extends GeoCoderEntity> i = result.iterator();
        assertGeoEntity(i.next(), "3016684", "", Type.INHABITED_PLACE, "?", 43.79,  7.52, "Garavan",            "Alpes-Maritimes", "Provence-Alpes-C\u00f4te d'Azur", "France");
        assertGeoEntity(i.next(), "6544017", "", Type.INHABITED_PLACE, "?", 43.79,  7.53, "grimaldi inferiore", "Provincia di Imperia", "Liguria", "Italy");
        assertGeoEntity(i.next(), "3175812", "", Type.INHABITED_PLACE, "?", 43.79,  7.54, "Grimaldi",           "Provincia di Imperia", "Liguria", "Italy");
        assertGeoEntity(i.next(), "3172545", "", Type.INHABITED_PLACE, "?", 43.80,  7.53, "Mortola Superiore",  "Provincia di Imperia", "Liguria", "Italy");
//        assertGeoEntity(i.next(), "2994497", 43.78,  7.50, "Menton",             "Département des Alpes-Maritimes", "Provence-Alpes-C\u00f4te d'Azur", "France");
        assertGeoEntity(i.next(), "3028380", "", Type.INHABITED_PLACE, "?", 43.80,  7.50, "Castellar",          "Castellar", "Arrondissement de Nice", "Alpes-Maritimes", "Provence-Alpes-C\u00f4te d'Azur", "France");
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Test
    public void shouldReturnValidDataForFranceRegions()
       throws NotFoundException
      {
        final GeoCoderEntity france = geoNamesProvider.findGeoEntityById("3017382");
        final Collection<? extends GeoCoderEntity> regions = france.findChildren().results();
        assertThat(regions.size(), is(26));
        final Iterator<? extends GeoCoderEntity> i = regions.iterator();
        assertGeoEntity(i.next(), "3038033", "C1", Type.ADMIN_DIVISION_1, "?", 48.50,  7.50, "Alsace",     "France");
        // TODO too lazy to test all...
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Test
    public void testFindCountryCode()
      {
        log("testFindCountryCode");
        assertEquals("IT", geoNamesProvider.findCountryCode(new Coordinate(45, 11)));
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Test
    public void testFactSheet()
       throws NotFoundException
      {
        log("testFactSheet");

        final SortedSet<Key<?>> keys = new TreeSet<Key<?>>(Arrays.asList(POPULATION, TIMEZONE));

        final FactSheet factSheetFrance = geoNamesProvider.findGeoEntityById("3017382").as(FactSheet);
        assertThat(factSheetFrance.toString(), factSheetFrance.getKeys(), is(keys));
        assertThat(factSheetFrance.get(POPULATION), is(64768389L));
        assertThat(factSheetFrance.get(TIMEZONE), is(TimeZone.getTimeZone("Europe/Paris")));

        final FactSheet factSheetBourgogne = geoNamesProvider.findGeoEntityById("3030967").as(FactSheet);
        assertThat(factSheetBourgogne.toString(), factSheetFrance.getKeys(), is(keys));
        assertThat(factSheetBourgogne.get(POPULATION), is(1633891L));
        assertThat(factSheetBourgogne.get(TIMEZONE), is(TimeZone.getTimeZone("Europe/Paris")));

        final FactSheet factSheetYonne = geoNamesProvider.findGeoEntityById("2967222").as(FactSheet);
        assertThat(factSheetYonne.toString(), factSheetFrance.getKeys(), is(keys));
        assertThat(factSheetYonne.get(POPULATION), is(341418L));
        assertThat(factSheetYonne.get(TIMEZONE), is(TimeZone.getTimeZone("Europe/Paris")));

        final FactSheet factSheetAuxerre = geoNamesProvider.findGeoEntityById("3035843").as(FactSheet);
        assertThat(factSheetAuxerre.toString(), factSheetFrance.getKeys(), is(keys));
        assertThat(factSheetAuxerre.get(POPULATION), is(41516L));
        assertThat(factSheetAuxerre.get(TIMEZONE), is(TimeZone.getTimeZone("Europe/Paris")));
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
//    @Test
//    public void findSubdivision()
//      {
//        log("findSubdivision");
//        GeoCoderEntity result = geoNamesProvider.findSubdivision(new Coordinate(45, 11));
//        assertNotNull(result);
//        assertEquals("Lombardy", result.as(Displayable).getDisplayName());
//        GeoCoderEntity country = result.getParent();
//        assertNotNull(country);
//        assertEquals("Italy", country.as(Displayable).getDisplayName());
//      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static void assertGeoEntity (GeoCoderEntity entity, 
                                         final String geonameId,
                                         final String code,
                                         final Type type,
                                         final String eblemMd5,
                                         final double latitude,
                                         final double longitude,
                                         final String ... names)
      throws NotFoundException
      {
        assertEquals(geonameId, entity.getId());
        assertNotNull(entity.getLookup().lookup(QName.class));
        final Coordinate coordinates = entity.getCoordinate();
        assertNotNull(coordinates);
        assertTrue("" + coordinates.getLatitude() + " -> " + latitude, Math.abs(coordinates.getLatitude() - latitude) < 0.005);
        assertTrue("" + coordinates.getLongitude() + " -> " + longitude, Math.abs(coordinates.getLongitude() - longitude) < 0.005);
        assertEquals(code, entity.getCode());
        assertEquals(type, entity.getType());

        final FactSheet factSheet = entity.as(FactSheet.class);
        assertThat("No FactSheet for " + entity , factSheet, is(notNullValue()));
        
        final EmblemFactory emblemFactory = entity.getLookup().lookup(GeoCoderEntity.EmblemFactory.class);
        assertThat("No EmblemFactory for " + entity , emblemFactory, is(notNullValue()));

        if (eblemMd5.equals(""))
          {
            try
              {
                emblemFactory.findEmblem(100);
                fail("NotFoundException expected");
              }
            catch (NotFoundException e)
              {
                // ok
              }
          }
        else
          {
            if (!"Åland".equals(names[0]))
              {
                final Image emblem = emblemFactory.findEmblem(100);
                assertThat(emblem, is(notNullValue()));
              }
          }

        for (int i = 0; i < names.length; i++)
          {
            assertNotNull(entity);
            assertEquals(names[i], entity.as(Displayable).getDisplayName());
            entity = entity.getParent();
          }
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static void log (final String message)
      {
        System.err.println(message);
      }
  }
