/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.viewer.dnd.geotag;

import javax.annotation.Nonnull;
import javax.inject.Inject;
import java.awt.datatransfer.DataFlavor;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetDropEvent;
import org.openide.util.Lookup;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.geo.Coordinate;
import it.tidalwave.geo.GeoTag;
import it.tidalwave.geo.viewer.FeatureManager;
import it.tidalwave.geo.viewer.GeoViewManager;

/***********************************************************************************************************************
 *
 * Accepts a drop event from an instance of {@link Lookup.Provider} associated to a {@link GeoTag}. The
 * {@code GeoTag} itself is set at the coordinate of the drop spot, then the object is added to the
 * {@link FeatureManager}.
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class GeoTaggingDropTarget extends DropTarget
  {
    private final static String CLASS = GeoTaggingDropTarget.class.getName();
    private final static Logger logger = Logger.getLogger(CLASS);
    
    @Inject
    private GeoViewManager geoViewManager;

    @Inject
    private FeatureManager featureManager;

    @Override
    public void drop (final @Nonnull DropTargetDropEvent event)
      {
        final Coordinate coordinate = geoViewManager.pointToCoordinate(event.getLocation());

        logger.fine("drop(%s)", event);
        logger.finer(">>>> coordinate at the drop spot: %s", coordinate);
        
        try
          {
            for (final DataFlavor flavor : event.getCurrentDataFlavors())
              {
                final Class<?> clazz = flavor.getRepresentationClass();
                logger.finest(">>>> trying flavor %s, class: %s...", flavor, clazz);

                if (Lookup.Provider.class.isAssignableFrom(clazz))
                  {
                    final Lookup.Provider lookupProvider = (Lookup.Provider)event.getTransferable().getTransferData(flavor);
                    final Lookup lookup = lookupProvider.getLookup();
                    final GeoTag geoTag = lookup.lookup(GeoTag.class);

                    logger.finest(">>>> geoTag: %s", geoTag);

                    if (geoTag != null)
                      {
                        // TODO: make this customizable
                        if (geoTag.isValid())
                          {
    //                        DialogRunner.showError("The object is already geotagged.", "Use 'Show in map' if you want to see the photo position.", "Error");
                            event.rejectDrop();
                          }
                        else
                          {
                            logger.finest(">>>> setting geoTag to %s", coordinate);
                            event.acceptDrop(event.getDropAction());
                            geoTag.setCoordinate(coordinate);
                            featureManager.addFeature(lookupProvider, coordinate);
                            event.dropComplete(true);
                          }

                        return; // don't do dropComplete(false);
                      }
                  }
              }
          }
        catch (Exception e)
          {
            logger.warning("Exception while managing drop: %s", e);
            logger.throwing(CLASS, "drop()", e);
          }

        event.dropComplete(false);
      }
  };

