/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
/***********************************************************************************************************************
 *
 * <p>This package provides a visual component for rendering 2D and 3D representations of geographic areas.
 *
 * <!-- img src="GeoViewerPresentation.png"/ -->
 *
 * This component does not expose a single interface for services; instead, it works as a factory where specific
 * services can be looked up:
 *
 * <pre>
 * GeoViewerPresentation geoViewerPresentation = GeoViewerPresentation.Locator.findGeoViewerPresentation();
 * GeoViewManager geoViewerManager = geoViewerPresentation.getLookup().lookup(GeoViewManager.class);
 * GeoViewProviderManager geoViewProviderManager = geoViewerPresentation.getLookup().lookup(GeoViewProviderManager.class);
 * FeatureManager featureManager = geoViewerPresentation.getLookup().lookup(FeatureManager.class);
 * </pre>
 *
 * The three exposed services are:
 *
 * <ul>
 * <li>{@link GeoViewManager}: it allows operations such as fitting the view to render the given {@link Coordinate}s
 *     etc.</li>
 * <li>{@link GeoViewProviderManager}: it allows to change the data provider (e.g. Microsoft Virtual Earth,
 *     OpenStreetMap) and the view mode (e.g. map, satellite, etc...).</li>
 * <li>{@link FeatureManager}: it allows to add and remove features on a view.</li>
 * </ul>
 * <p>
 *
 * <h3>Injectable roles</h3>
 *
 * <p>The implementation of this component is delegated to a set of small and cohese classes named "roles". Each
 * one has got a single reponsibility and usually can be replaced independently of the others. Each role has a
 * symbolic name, which is part of the public API, and an implementation class.<br/>
 * Most roles have a default implementation; some are mandatory (i.e. their absence breaks the
 * {@code GeoViewerPresentation}}, while others are optional and can be added by the programmer.</p>
 *
 * <table border="1" cellpadding="4" cellspacing="0">
 * <tr><th>Role</th><th>Description</th><th>Implementation class</th><th>Description</th></tr>
 * <tr>
 * <td rowspan="2">{@code CacheLocationProvider}<br/>(mandatory)</td>
 * <td rowspan="2">Provides the path of a folder where a cache of geographical data (e.g. tiles).</td>
 * <td>default</td>
 * <td>The cache is under $HOME/.ForceTen/cache</td>
 * </tr>
 * <tr>
 * <td>custom roles</td>
 * <td>Can be implemented by subclassing
 *     {@link it.tidalwave.geo.viewer.role.CacheLocationProvider CacheLocationProvider}.
 * </td>
 * </tr>
 *
 * <tr>
 * <td rowspan="2">{@code InitialLocationProvider}<br/>(mandatory)</td>
 * <td rowspan="2">Provides the initial coordinate where the geographic view will be centered at.</td>
 * <td>{@link it.tidalwave.geo.viewer.role.ZeroZeroInitialLocationProvider ZeroZeroInitialLocationProvider}<br/>default</td>
 * <td>The initial location is at coordinate (0; 0)</td>
 * </tr>
 * <tr>
 * <td>custom roles</td>
 * <td>Can be implemented by subclassing
 *     {@link it.tidalwave.geo.viewer.role.InitialLocationProvider InitialLocationProvider}.
 * </td>
 * </tr>
 *
 * <tr>
 * <td rowspan="2">{@code UI}<br/>(mandatory)</td>
 * <td rowspan="2">Provides the visual component. While it is obvious that such a component renders the geographical
 *                 view, this role can be customized to add controls in different places.</td>
 * <td>default</td>
 * <td>The visual component include a toolbar at the bottom of the UI with controls for selecting the
 *     {@link GeoViewProviderManager}, the rendering mode, a customizable button bar and a label with renders the
 *     coordinates under the current mouse position.
 * </td>
 * </tr>
 * <tr>
 * <td>custom roles</td>
 * <td>Can be implemented by subclassing
 *     {@link it.tidalwave.geo.viewer.role.GeoViewerUI GeoViewerUI}.
 * </td>
 * </tr>
 *
 * <tr>
 * <td rowspan="1">{@code ContextListener}<br/>(optional)</td>
 * <td rowspan="1">Listens for the {@link org.netbeans.platformx.eventbus.api.EventBus EventBus} and fits the
 *     geographical view when a {@link it.tidalwave.geo.geocoding.GeoEntitySet GeoEntitySet} or a
 *     {@link it.tidalwave.geo.Sector Sector} is published.</td>
 * <td></td>
 * <td></td>
 * </tr>
 * </table>
 * <br/>
 *
 *
 * <h3>Customization and configuration</h3>
 *
 * Configuration is performed by means of {@code layer.xml}. The following is the default configuration of
 * {@code GeoViewerPresentation}:
 *
 * <pre>
 * &lt;?xml version="1.0" encoding="UTF-8"?&gt;
 * &lt;!DOCTYPE filesystem PUBLIC "-//NetBeans//DTD Filesystem 1.1//EN" "http://www.netbeans.org/dtds/filesystem-1_1.dtd"&gt;
 * &lt;filesystem&gt;
 *     &lt;folder name="Roles"&gt;
 *         &lt;folder name="it.tidalwave.geo.explorer.GeoViewerPresentation"&gt;
 *             &lt;file name="GeoViewProviderManager.instance"&gt;
 *                 ...
 *             &lt;/file&gt;
 *             &lt;file name="GeoViewManager.instance"&gt;
 *                 ...
 *             &lt;/file&gt;
 *             &lt;file name="InitialLocationProvider.instance"&gt;
 *                 ...
 *             &lt;/file&gt;
 *             &lt;file name="FeatureManager.instance"&gt;
 *                 ...
 *             &lt;/file&gt;
 *             &lt;file name="ContextListener.instance"&gt;
 *                 ...
 *             &lt;/file&gt;
 *         &lt;/folder&gt;
 *     &lt;/folder&gt;
 * ...
 * &lt;/filesystem&gt;
 * </pre>
 *
 * In order to configure {@code GeoViewerPresentation} you need to create a specific NetBeans Platform module with
 * a dependency on this module and a {@code layer.xml}.
 * To replace a default role, the relevant definition must be overridden. For instance, this configuration replaces
 * the "(0;0) initial location provider" strategy with another:
 *
 * <pre>
 * &lt;?xml version="1.0" encoding="UTF-8"?&gt;
 * &lt;!DOCTYPE filesystem PUBLIC "-//NetBeans//DTD Filesystem 1.1//EN" "http://www.netbeans.org/dtds/filesystem-1_1.dtd"&gt;
 * &lt;filesystem&gt;
 *     &lt;folder name="Roles"&gt;
 *         &lt;folder name="it.tidalwave.geo.explorer.GeoViewerPresentation"&gt;
 *             <b>&lt;file name="InitialLocationProvider.instance"&gt;
 *                 &lt;attr name="instanceClass" stringvalue="my.own.InitialLocationProvider"/&gt;
 *             &lt;/file&gt;</b>
 *         &lt;/folder&gt;
 *     &lt;/folder&gt;
 * &lt;/filesystem&gt;
 * </pre>
 *
 * It is also possible to disable a default role without replacing it; the following configuration disables the
 * "context listener" (for instance because you want to manually control the geographical view):
 *
 * <pre>
 * &lt;?xml version="1.0" encoding="UTF-8"?&gt;
 * &lt;!DOCTYPE filesystem PUBLIC "-//NetBeans//DTD Filesystem 1.1//EN" "http://www.netbeans.org/dtds/filesystem-1_1.dtd"&gt;
 * &lt;filesystem&gt;
 *     &lt;folder name="Roles"&gt;
 *         &lt;folder name="it.tidalwave.geo.explorer.GeoViewerPresentation"&gt;
 *             <b>&lt;file name="ContextListener.instance_hidden"/&gt;</b>
 *         &lt;/folder&gt;
 *     &lt;/folder&gt;
 * &lt;/filesystem&gt;
 * </pre>
 *
 * Last but not least, it is also possible to add new roles:
 *
 * <pre>
 * &lt;?xml version="1.0" encoding="UTF-8"?&gt;
 * &lt;!DOCTYPE filesystem PUBLIC "-//NetBeans//DTD Filesystem 1.1//EN" "http://www.netbeans.org/dtds/filesystem-1_1.dtd"&gt;
 * &lt;filesystem&gt;
 *     &lt;folder name="Roles"&gt;
 *         &lt;folder name="it.tidalwave.geo.explorer.GeoViewerPresentation"&gt;
 *             <b>&lt;file name="MyNewRole.instance"&gt;
 *                 &lt;attr name="instanceClass" stringvalue="my.newrole.Implementation"/&gt;
 *             &lt;/file&gt;</b>
 *         &lt;/folder&gt;
 *     &lt;/folder&gt;
 * &lt;/filesystem&gt;
 * </pre>
 *
 *
 * <h3>Default state</h3>
 *
 * This component is placed by default in the "center" mode and it's not set visible by default. If you want to be
 * visible by default, use this customization fragment of {@code layer.xml}:
 *
 * <pre>
 * &lt;?xml version="1.0" encoding="UTF-8"?&gt;
 * &lt;!DOCTYPE filesystem PUBLIC "-//NetBeans//DTD Filesystem 1.1//EN" "http://www.netbeans.org/dtds/filesystem-1_1.dtd"&gt;
 * &lt;filesystem&gt;
 *     &lt;folder name="Windows2"&gt;
 *         &lt;folder name="Modes"&gt;
 *             &lt;folder name="center"&gt;
 *                 &lt;file name="it.tidalwave.geo.viewer.GeoViewerPresentation.wstcref" url="GeoViewerPresentationWstcref.xml"/&gt;
 *             &lt;/folder&gt;
 *         &lt;/folder&gt;
 *     &lt;/folder&gt;
 * &lt;/filesystem&gt;
 * </pre>
 *
 * with {@code GeoViewerPresentationWstcref.xml} being:
 *
 * <pre>
 * &lt;?xml version="1.0" encoding="UTF-8"?&gt;
 * &lt;!DOCTYPE tc-ref PUBLIC "-//NetBeans//DTD Top Component in Mode Properties 2.0//EN" "http://www.netbeans.org/dtds/tc-ref2_0.dtd"&gt;
 * &lt;tc-ref version="2.0" &gt;
 *     &lt;module name="it.tidalwave.geo.viewer" spec="1.0"/&gt;
 *     &lt;tc-id id="it.tidalwave.geo.viewer.GeoViewerPresentation"/&gt;
 *     &lt;state opened="true"/&gt;
 * &lt;/tc-ref&gt;
 * </pre>
 *
 *
 * <h3>Predefined actions</h3>
 *
 * An action to toggle the visibility status of the {@link GeoViewerPresentation} is automatically installed under the
 * "Window" menu. The name of the file representing the action, that can be used to customize the position or remove the
 * action completely, is
 *
 * <pre>
 * "Actions/Window/it-tidalwave-geo-viewer-impl-action-GeoViewerAction.instance"
 * </pre>
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.viewer;

