/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.viewer.spi.jxmapviewer;

import javax.annotation.Nonnull;
import javax.annotation.PostConstruct;
import javax.inject.Inject;
import java.awt.Component;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.awt.event.ActionEvent;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.ButtonGroup;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JPopupMenu;
import org.jdesktop.swingx.JXMapViewer;
import org.jdesktop.swingx.mapviewer.DefaultTileFactory;
import org.jdesktop.swingx.mapviewer.GeoPosition;
import org.jdesktop.swingx.mapviewer.TileCache;
import org.jdesktop.swingx.mapviewer.TileFactory;
import org.jdesktop.swingx.mapviewer.TileFactoryInfo;
import org.openide.util.Lookup;
import org.openide.util.lookup.ServiceProvider;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.role.Displayable;
import it.tidalwave.geo.Coordinate;
import it.tidalwave.geo.Sector;
import it.tidalwave.geo.viewer.spi.GeoViewProvider;
import it.tidalwave.geo.viewer.spi.GeoViewProviderSupport;
import it.tidalwave.geo.viewer.spi.ResourceCache;
import it.tidalwave.geo.viewer.spi.TileProviderDescriptor;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@ServiceProvider(service=GeoViewProvider.class)
public class JXMapViewerGeoViewProvider extends GeoViewProviderSupport
  {
    private static final String CLASS = JXMapViewerGeoViewProvider.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);

    @Inject
    private JXMapViewerGeoViewSupport mapViewerSupport;

    @Inject
    private ResourceCache resourceCache;

    @Nonnull
    private JXMapViewer mapViewer;

    private TileFactoryInfo tileFactoryInfo;

    private final JPopupMenu optionsMenu = new JPopupMenu();

    private final ButtonGroup buttonGroup = new ButtonGroup();

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public JXMapViewerGeoViewProvider()
      {
        super("2D", 7); // FIXME
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    @Nonnull
    public Point coordinatesToPoint (final @Nonnull Coordinate coordinates)
      {
        final Point2D point = mapViewer.getTileFactory().geoToPixel(CoordinateConverter.toGeoPosition(coordinates),
                                                                    mapViewer.getZoom());
        final Rectangle viewportBounds = mapViewer.getViewportBounds();
        return new Point((int)point.getX() - (int)viewportBounds.getX(), (int)point.getY() - (int)viewportBounds.getY());
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    @Nonnull
    public Coordinate pointToCoordinates (final @Nonnull Point point)
      {
        return pointToCoordinates(point, mapViewer.getZoom());
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public Coordinate pointToCoordinates (final @Nonnull Point point, final int zoom)
      {
        final Rectangle viewportBounds = mapViewer.getViewportBounds();
        final Point2D.Double point2 = new Point2D.Double(point.getX() + viewportBounds.getX(), point.getY() + viewportBounds.getY());
        return CoordinateConverter.fromGeoPosition(mapViewer.getTileFactory().pixelToGeo(point2, zoom));
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public void activated()
      {
        super.activated();
        final NavigationPane navigationPane = mapViewerSupport.getNavigationPane();
        navigationPane.setMaxZoom(getMaximumZoomLevel());
        navigationPane.setMinZoom(getMinimumZoomLevel());
        navigationPane.setZoom(getZoom());
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public final int getMinimumZoomLevel()
      {
        return tileFactoryInfo.getMinimumZoomLevel();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public final int getMaximumZoomLevel()
      {
        return tileFactoryInfo.getMaximumZoomLevel();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    protected TileFactory getTileFactory()
      {
        final DefaultTileFactory tileFactory = new DefaultTileFactory(tileFactoryInfo);
        final TileCache tileCache = tileFactory.getTileCache();
        tileFactory.setTileCache(new TileCacheDecorator(tileCache, resourceCache));
        return tileFactory;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    @Nonnull
    public final Component getRendererComponent()
      {
        return mapViewerSupport.getComponent();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public final void setZoom (final int zoom)
      {
        logger.fine("setZoom(%d)", zoom);
        mapViewer.setZoom(Math.max(getMinimumZoomLevel(), Math.min(getMaximumZoomLevel(), zoom)));
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public final int getZoom()
      {
        return mapViewer.getZoom();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    @Nonnull
    public final Coordinate getCenterPosition()
      {
        return CoordinateConverter.fromGeoPosition(mapViewer.getCenterPosition());
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public final void setCenterPosition (final @Nonnull Coordinate centerPosition)
      {
        logger.fine("setCenterPosition(%s)", centerPosition);
        mapViewer.setCenterPosition(CoordinateConverter.toGeoPosition(centerPosition));
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public final void pan (final double x, final double y)
      {
        logger.fine("pan(%d, %d)", x, y);
        final GeoPosition center = mapViewer.getAddressLocation();
        mapViewer.setAddressLocation(new GeoPosition(center.getLongitude() + x * 5, center.getLatitude() + y * 5)); // FIXME * 5
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public void fitView (final @Nonnull Sector sector)
      {
        // Since the "zoom" value doesn't have an absolute value, the computation is for a relative zoom to apply.
        logger.fine("fitView(%s)", sector);
        final Sector bounds = getSector();
        final double deltaSector = sector.getMaxDelta();
        final double deltaBounds = bounds.getMaxDelta();
        logger.finest(">>>> bounds: %s, deltaSector: %s, deltaBounds: %s", bounds, deltaSector, deltaBounds);
        final double scale = deltaSector / deltaBounds;
        logger.finest(">>>> scale: %f", scale);
        final int deltaZoom = (int)Math.ceil(Math.log(scale) / Math.log(2.0));
        logger.finest(">>>> deltaZoom: %d", deltaZoom);
        setCenterPosition(sector.getCenter());
        setZoom(getZoom() + deltaZoom);
     }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void setTileProviderDescriptor (final @Nonnull TileProviderDescriptor tileProviderDescriptor)
      {
        super.setGeoDataProviderDescriptor(tileProviderDescriptor);
        tileFactoryInfo = new TileFactoryInfoAdapter(tileProviderDescriptor);

        if (mapViewer != null)
          {
            final GeoPosition centerPosition = mapViewer.getCenterPosition();
            mapViewer.setTileFactory(getTileFactory());
            mapViewer.setCenterPosition(centerPosition);
            mapViewer.repaint();
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public JPopupMenu getOptionsMenu()
      {
        return optionsMenu;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @PostConstruct
    @SuppressWarnings("PMD.UnusedPrivateMethod") @edu.umd.cs.findbugs.annotations.SuppressWarnings("UPM_UNCALLED_PRIVATE_METHOD")
    private void initialize()
      {
        mapViewer = mapViewerSupport.getMapViewer();
        boolean first = true;

        for (final TileProviderDescriptor tileProviderDescriptor : Lookup.getDefault().lookupAll(TileProviderDescriptor.class))
          {
            final Action action = new AbstractAction()
              {
                @Override
                public void actionPerformed (final @Nonnull ActionEvent event)
                  {
                    setTileProviderDescriptor(tileProviderDescriptor);
                  }
              };

            action.putValue(Action.NAME, tileProviderDescriptor.as(Displayable.class).getDisplayName());
            final JCheckBoxMenuItem cbMenuItem = new JCheckBoxMenuItem(action);
            buttonGroup.add(cbMenuItem);
            optionsMenu.add(cbMenuItem);

            if (first)
              {
                first = false;
                cbMenuItem.setState(true);
                setTileProviderDescriptor(tileProviderDescriptor);
              }
          }
      }

    /*******************************************************************************************************************
     *
     * Returns the currently shown sector.
     *
     ******************************************************************************************************************/
    @Nonnull
    private Sector getSector()
      {
        final Point topLeft = new Point(0, 0);
        final Point bottomRight = new Point(mapViewer.getBounds().width - 1, mapViewer.getBounds().height - 1);
        return new Sector(pointToCoordinates(topLeft), pointToCoordinates(bottomRight));
      }
  }
