/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.viewer.spi.jxmapviewer;

import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import javax.annotation.PostConstruct;
import javax.inject.Inject;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.dnd.DropTarget;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.geom.Point2D;
import javax.swing.BorderFactory;
import javax.swing.JComponent;
import org.jdesktop.beansbinding.AutoBinding.UpdateStrategy;
import org.jdesktop.beansbinding.BeanProperty;
import org.jdesktop.beansbinding.BindingGroup;
import org.jdesktop.beansbinding.Bindings;
import org.jdesktop.observablecollections.ObservableList;
import org.jdesktop.observablecollections.ObservableListListener;
import org.jdesktop.swingx.JXMapViewer;
import org.jdesktop.swingx.mapviewer.GeoPosition;
import org.openide.util.Lookup;
import org.openide.util.lookup.Lookups;
import org.openide.util.lookup.ProxyLookup;
import org.openide.nodes.Children;
import org.openide.nodes.FilterNode;
import org.openide.nodes.Node;
import org.netbeans.api.visual.model.ObjectScene;
import org.netbeans.api.visual.widget.Widget;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.netbeans.nodes.role.RootNodeProvider;
import it.tidalwave.netbeans.visual.LayeredSceneView;
import it.tidalwave.netbeans.visual.LayeredSceneView.LocationProvider;
import it.tidalwave.netbeans.visual.NodeScene;
import it.tidalwave.netbeans.visual.NodeWidgetFactory;
import it.tidalwave.netbeans.visual.action.BringToFrontAction;
import it.tidalwave.geo.Coordinate;
import it.tidalwave.geo.GeoTag;
import it.tidalwave.geo.geocoding.GeoCoderEntity;
import it.tidalwave.geo.viewer.GeoViewManager;
import it.tidalwave.geo.viewer.spi.GeoViewProvider.CoordinateTracker;
import it.tidalwave.geo.viewer.spi.feature.FeatureManagerSpi;
import it.tidalwave.geo.viewer.spi.jxmapviewer.widget.GeoConnectingWidget;
import it.tidalwave.geo.viewer.spi.jxmapviewer.widget.GeoPinWidget;
import it.tidalwave.geo.viewer.spi.jxmapviewer.widget.OffsetWidget;
import org.netbeans.platformx.inject.api.Injector;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class JXMapViewerGeoViewSupport implements RootNodeProvider
  {
    private static final String CLASS = JXMapViewerGeoViewSupport.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);

    @Inject
    private Lookup lookup;

    @Inject
    private CoordinateTracker coordinateTracker;

    @Inject
    private FeatureManagerSpi featureManagerSpi;

    @Inject
    private GeoViewManager geoViewManager;

    @Inject
    private DropTarget dropTarget;

    private LayeredSceneView layeredSceneView;

    private final JXMapViewer mapViewer = new JXMapViewer();

    private final NavigationPane navigationPane = new NavigationPane();

    private final BindingGroup bindings = new BindingGroup();

    private final GeoTaggingDragBehaviour dragBehaviour = new GeoTaggingDragBehaviour();

    private final NodeScene scene = new NodeScene();

    // FIXME: declare in layer.xml
    private final static BringToFrontAction bringToFrontAction = new BringToFrontAction();
    // FIXME: declare in layer.xml

    /*******************************************************************************************************************
     *
     * This factory decorates the original {@link NodeWidgetFactory} that is found in the {@link GeoCoderEntity}'s
     * {@link Lookup} and adds the {@link GeoPinWidget} and the {@link GeoConnectingWidget}.
     *
     ******************************************************************************************************************/
    private static class NodeWidgetFactoryDecorator implements NodeWidgetFactory
      {
        @Nonnull
        private final NodeWidgetFactory delegate;

        @Nonnull
        private final NodeScene.DragBehaviour dragBehaviour;

        public NodeWidgetFactoryDecorator (final @Nonnull NodeWidgetFactory delegate,
                                           final @Nonnull NodeScene.DragBehaviour dragBehaviour)
          {
            this.delegate = delegate;
            this.dragBehaviour = dragBehaviour;
          }

        @Override @Nonnull
        public Collection<Widget> createWidgets (final @Nonnull ObjectScene scene, final @Nonnull Node node)
          {
            logger.fine("createWidgets(%s, %s)", scene, node);
            
            final Widget widget = delegate.createWidgets(scene, node).iterator().next();
            final Widget pinWidget = new GeoPinWidget(scene, widget, dragBehaviour);
            final GeoConnectingWidget connectingWidget = new GeoConnectingWidget(scene, widget, pinWidget);
            final OffsetWidget offsetWidget = new OffsetWidget(scene, widget, pinWidget);

            pinWidget.getActions().addAction(bringToFrontAction);
            pinWidget.addDependency(offsetWidget);
            offsetWidget.getActions().addAction(bringToFrontAction);

            return Arrays.asList(pinWidget, offsetWidget, connectingWidget);
          }
      }

    /*******************************************************************************************************************
     *
     * This class keeps the Widgets on the Scene at the right position. 
     *
     ******************************************************************************************************************/
    private final LocationProvider<Node> locationProvider = new LocationProvider<Node>()
      {
        @Override @Nonnull
        public Point findLocation (final @Nonnull Node node)
          {
            if (dragBehaviour.getDraggedNode() == node)
              {
                return null;
              }

            final GeoTag geoTag = node.getLookup().lookup(GeoTag.class);
            final Coordinate coordinate = (geoTag != null) ? geoTag.getCoordinate() : node.getLookup().lookup(Coordinate.class);
            assert coordinate != null;

            return geoViewManager.coordinateToPoint(coordinate);
          }
      };

    /*******************************************************************************************************************
     *
     * Tracks mouse motion and shows the current coordinates.
     *
     ******************************************************************************************************************/
    private final MouseMotionAdapter coordinateTrackerListener = new MouseMotionAdapter()
      {
        private final Point2D.Float point = new Point2D.Float();

        @Override
        public void mouseMoved (final @Nonnull MouseEvent event)
          {
            final Rectangle viewportBounds = mapViewer.getViewportBounds();
            point.setLocation(event.getX() + viewportBounds.getX(), event.getY() + viewportBounds.getY());
            final GeoPosition position = mapViewer.getTileFactory().pixelToGeo(point, mapViewer.getZoom()); // FIXME: convertPointToGeoPosition(point)?
            coordinateTracker.coordinatesChanged(new Coordinate(position.getLatitude(), position.getLongitude()));
          }
      };

    /*******************************************************************************************************************
     *
     * Tracks changes in features and create / removes Widgets as features gets added and removed.
     *
     ******************************************************************************************************************/
    private final ObservableListListener featureListener = new ObservableListListener()
      {
        @Override
        public void listElementsAdded (final @Nonnull ObservableList features,
                                       final @Nonnegative int from,
                                       final @Nonnegative int count)
          {
            logger.finer("listElementsAdded(%s, %d, %d)", features, from, count);
            
            for (int i = from; i < from + count; i++)
              {
                final Object feature = features.get(i);
                final Node node = featureManagerSpi.findNode(feature);
                logger.finest(">>>> feature: %s node: %s", feature, node);
                final Lookup originalLookup = node.getLookup();
                final NodeWidgetFactory nodeWidgetFactory = originalLookup.lookup(NodeWidgetFactory.class);
                assert nodeWidgetFactory != null : "No NodeWidgetFactory in node " + node + ": " + originalLookup.lookupAll(Object.class);
                final Lookup filteredLookup = Lookups.exclude(originalLookup, NodeWidgetFactory.class);
                final Lookup extraLookup = Lookups.fixed(new NodeWidgetFactoryDecorator(nodeWidgetFactory, dragBehaviour));
                scene.addNode(new FilterNode(node, Children.LEAF, new ProxyLookup(filteredLookup, extraLookup)));
              }
          }

        @Override
        public void listElementsRemoved (final @Nonnull ObservableList features,
                                         final @Nonnegative int index,
                                         final @Nonnull List removedFeatures)
          {
            logger.finer("listElementsRemoved(%s, %d, %s)", features, index, removedFeatures);

            for (final Object feature : removedFeatures)
              {
                final Node node = featureManagerSpi.findNode(feature);
                scene.removeObject(node);
//                sceneFacade.removeWidget(...); // FIXME
              }
          }

        @Override
        public void listElementReplaced (final @Nonnull ObservableList features,
                                         final @Nonnegative int i,
                                         final @Nonnull Object feature)
          {
          }

        @Override
        public void listElementPropertyChanged (final @Nonnull ObservableList ol,
                                                final @Nonnegative int i)
          {
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public JComponent getComponent()
      {
        return layeredSceneView;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public JXMapViewer getMapViewer()
      {
        return mapViewer;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public NavigationPane getNavigationPane()
      {
        return navigationPane;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Node getRootNode()
      {
        return scene.getRootNode();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @PostConstruct
    @SuppressWarnings("PMD.UnusedPrivateMethod") @edu.umd.cs.findbugs.annotations.SuppressWarnings("UPM_UNCALLED_PRIVATE_METHOD")
    private void initialize()
      {
        featureManagerSpi.getFeatures().addObservableListListener(featureListener);

        mapViewer.setRecenterOnClickEnabled(true);
        mapViewer.addMouseMotionListener(coordinateTrackerListener);
        mapViewer.setBorder(BorderFactory.createEmptyBorder());
        mapViewer.setOverlayPainter(lookup.lookup(JXMapViewerFeaturePainter.class));

        bindings.addBinding(Bindings.createAutoBinding(UpdateStrategy.READ_WRITE,
                                                       mapViewer, BeanProperty.create("zoom"),
                                                       navigationPane, BeanProperty.create(NavigationPane.PROP_ZOOM)));
        bindings.bind();

        layeredSceneView = new LayeredSceneView(scene, locationProvider, mapViewer, navigationPane);
        layeredSceneView.addRevalidateTrigger(mapViewer, "zoom", "centerPosition");

        if (dropTarget != null)
          {
            layeredSceneView.setDropTarget(dropTarget); // FIXME
//            scene.getView().setDropTarget(dropTarget); 
          }

        Injector.getDefault().inject(dragBehaviour, lookup); // FIXME: temporary, make dragBehaviour injected instead
      }
  }
