/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.viewer.spi.microsoftvirtualearth;

import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import java.util.Arrays;
import java.util.Collections;
import java.util.Locale;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;
import java.awt.Dimension;
import java.net.MalformedURLException;
import java.net.URL;
import it.tidalwave.role.Displayable;
import it.tidalwave.geo.viewer.spi.TileProviderDescriptor;
import it.tidalwave.role.LocalizedDisplayable;

/***********************************************************************************************************************
 *
 * See http://www.viavirtualearth.com/vve/Articles/RollYourOwnTileServer.ashx
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public abstract class MicrosoftVirtualEarthTileProviderDescriptor extends TileProviderDescriptor
  {
    private final static int MIN_ZOOM_LEVEL = 2;

    private final static int MAX_ZOOM_LEVEL = 17;

    private final static int TOP_ZOOM_LEVEL = 19;

    private final static int INITIAL_ZOOM_LEVEL = 12;

    @Nonnull
    private final String type;

    @Nonnull
    private final String extension;

    @Nonnull
    private final String typeDisplayName;

    private final LocalizedDisplayable displayable = new LocalizedDisplayable()
      {
        @Override @Nonnull
        public String getDisplayName()
          {
            return getDisplayName(Locale.ENGLISH);
          }

        @Override @Nonnull
        public String getDisplayName (final @Nonnull Locale locale)
          {
            return "Microsoft Virtual Earth (" + typeDisplayName + ")";
          }

        @Override @Nonnull
        public Map<Locale, String> getDisplayNames()
          {
            return Collections.singletonMap(Locale.ENGLISH, getDisplayName());
          }

        @Override @Nonnull
        public SortedSet<Locale> getLocales()
          {
            return new TreeSet<Locale>(Arrays.asList(Locale.ENGLISH));
          }
      };

    protected MicrosoftVirtualEarthTileProviderDescriptor (final @Nonnull String type,
                                                           final @Nonnull String extension,
                                                           final @Nonnull String typeDisplayName)
      {
        this.type = type;
        this.extension = extension;
        this.typeDisplayName = typeDisplayName;
      }

    @Override @Nonnull
    public Dimension getTileSize()
      {
        return new Dimension(256, 256);
      }

    @Override @Nonnegative
    public int getLevelCount()
      {
        return 16;
      }

    @Override @Nonnegative
    public int getMinZoomLevel()
      {
        return MIN_ZOOM_LEVEL;
      }

    @Override @Nonnegative
    public int getMaxZoomLevel()
      {
        return MAX_ZOOM_LEVEL;
      }

    @Override @Nonnegative
    public int getTopZoomLevel()
      {
        return TOP_ZOOM_LEVEL;
      }

    @Override @Nonnegative
    public int getInitialZoomLevel()
      {
        return INITIAL_ZOOM_LEVEL;
      }

    @Override @Nonnull
    public HorizontalLayout getHorizontalLayout()
      {
        return HorizontalLayout.LEFT_TO_RIGHT;
      }

    @Override @Nonnull
    public VerticalLayout getVerticalLayout()
      {
        return VerticalLayout.BOTTOM_TO_TOP;
      }

    @Override @Nonnull
    public String getDataCacheName()
      {
        return "MVE/" + type;
      }

    @Override @Nonnull
    public URL getTileUrl (final @Nonnegative int level, final int column, final int row)
      throws MalformedURLException
      {
        final String quad = tileToQuadKey(column, row, level);
        return new URL(String.format("http://%s%s.ortho.tiles.virtualearth.net/tiles/%s%s.%s?g=1",
                                     type, quad.charAt(quad.length() - 1), type, quad, extension));
      }

    @Nonnull
    protected static String tileToQuadKey (final int column, final int row, final @Nonnegative int level)
      {
        final StringBuilder buffer = new StringBuilder();

        for (int i = level - 1; i >= 0; i--)
          {
            final int mask = 1 << i;
            int cell = 0;

            if ((column & mask) != 0)
              {
                cell++;
              }

            if ((row & mask) != 0)
              {
                cell += 2;
              }

            buffer.append(cell);
          }

        return buffer.toString();
      }

    @Override @Nonnull
    protected Displayable getDisplayable()
      {
        return displayable;
      }
  }
