/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.viewer.worldwindprovider.impl;

import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import javax.inject.Inject;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.ActionEvent;
import javax.swing.AbstractAction;
import javax.swing.JPanel;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JPopupMenu;
import gov.nasa.worldwind.Model;
import gov.nasa.worldwind.View;
import gov.nasa.worldwind.WorldWind;
import gov.nasa.worldwind.WorldWindow;
import gov.nasa.worldwind.avlist.AVKey;
import gov.nasa.worldwind.event.PositionEvent;
import gov.nasa.worldwind.event.PositionListener;
import gov.nasa.worldwind.geom.Angle;
import gov.nasa.worldwind.geom.Position;
import gov.nasa.worldwind.globes.Globe;
import gov.nasa.worldwind.layers.CompassLayer;
import gov.nasa.worldwind.layers.RenderableLayer;
import gov.nasa.worldwind.layers.TiledImageLayer;
import gov.nasa.worldwind.layers.FogLayer;
import gov.nasa.worldwind.layers.Layer;
import gov.nasa.worldwind.layers.LayerList;
import gov.nasa.worldwind.layers.ScalebarLayer;
import gov.nasa.worldwind.layers.SkyGradientLayer;
import gov.nasa.worldwind.layers.StarsLayer;
import gov.nasa.worldwind.layers.WorldMapLayer;
import gov.nasa.worldwind.layers.Earth.BMNGOneImage;
import gov.nasa.worldwind.layers.Earth.BMNGSurfaceLayer;
import gov.nasa.worldwind.layers.Earth.CountryBoundariesLayer;
import gov.nasa.worldwind.layers.Earth.EarthNASAPlaceNameLayer;
import gov.nasa.worldwind.layers.Earth.LandsatI3;
import gov.nasa.worldwind.layers.Earth.USGSDigitalOrtho;
import gov.nasa.worldwind.layers.Earth.USGSUrbanAreaOrtho;
import gov.nasa.worldwind.render.Renderable;
import gov.nasa.worldwind.view.FlyToOrbitViewStateIterator;
import gov.nasa.worldwind.view.OrbitView;
import org.openide.util.NbBundle;
import org.openide.util.Lookup;
import org.openide.util.lookup.ServiceProvider;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.netbeans.dialog.DialogRunner;
import it.tidalwave.geo.Coordinate;
import it.tidalwave.geo.Sector;
import it.tidalwave.geo.viewer.spi.GeoViewProviderSupport;
import it.tidalwave.geo.viewer.ZoomLevel;
import it.tidalwave.geo.viewer.spi.GeoViewProvider;
import it.tidalwave.geo.viewer.spi.TileProviderDescriptor;
import it.tidalwave.geo.viewer.worldwindprovider.spi.LayerProvider;
import it.tidalwave.geo.viewer.worldwindprovider.impl.layer.GridLayer;
import it.tidalwave.geo.viewer.worldwindprovider.impl.layer.PathLayer;
import it.tidalwave.geo.viewer.worldwindprovider.impl.layer.TiledImageLayerProvider;
import it.tidalwave.geo.viewer.worldwindprovider.impl.layer.WaypointLayer;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@ServiceProvider(service=GeoViewProvider.class)
public class NASAWorldWindProvider extends GeoViewProviderSupport
  {
    private final static String CLASS = NASAWorldWindProvider.class.getName();    
    private final static Logger logger = Logger.getLogger(CLASS);
    
//    private final static int TOP_ZOOM_LEVEL = 19;
    
    private final static int MAX_ZOOM_LEVEL = 17;
    
    private final static int MIN_ZOOM_LEVEL = 2;
    
    private final static int INITIAL_ZOOM_LEVEL = 12;
    
    @Inject
    private WorldWindow worldWindow;
    
    private final JPanel thePanel = new JPanel(new BorderLayout());
    
//    private Coordinate centerPosition = new Coordinate(0, 0);
    
    private final LayerList layerList = new LayerList();
      
    private final JPopupMenu optionsMenu = new JPopupMenu();

    private long centerAnimationDelay = 2000;

    @Inject
    private Lookup lookup;
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private final PositionListener positionListener = new PositionListener() 
      {
        @Override
        public void moved (final @Nonnull PositionEvent positionEvent)
          {
            if (coordinatesTracker != null)
              {
                final Position position = positionEvent.getPosition();
                coordinatesTracker.coordinatesChanged((position == null) ? null : toCoordinate(position));
              }
          }
      };
      
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    class LayerAction extends AbstractAction
      {
        @Nonnull
        private final Layer layer;

        public LayerAction (final @Nonnull Layer layer, final boolean selected)
          {
            super(layer.getName());
            this.layer = layer;
            this.layer.setEnabled(selected);
          }

        @Override
        public void actionPerformed (final @Nonnull ActionEvent actionEvent)
          {
            layer.setEnabled(((JCheckBoxMenuItem)actionEvent.getSource()).isSelected());
            worldWindow.redraw();
          }
        
        public boolean isSelected()
          {
            return layer.isEnabled();
          }
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    static class GenericLayer extends RenderableLayer
      {
        public GenericLayer (final @Nonnull String name, final @Nonnull Renderable renderable)
          {
            setName(name);
            addRenderable(renderable);
          }
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public NASAWorldWindProvider()
      {
        super("3D", INITIAL_ZOOM_LEVEL);
        registerZoomLevel(ZoomLevel.STREET, 2);
        registerZoomLevel(ZoomLevel.TOWN, 5);
        registerZoomLevel(ZoomLevel.PROVINCE, 8);
        registerZoomLevel(ZoomLevel.REGION, 10);
        registerZoomLevel(ZoomLevel.COUNTRY, 12);
        registerZoomLevel(ZoomLevel.CONTINENT, 14);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void initResources()
      {
        logger.fine("initResources()");
        
        try
          {
            final LayerAction[] layerActions1 = new LayerAction[] 
              {
                new LayerAction(new StarsLayer(), true),
                new LayerAction(new FogLayer(), true),
                new LayerAction(new SkyGradientLayer(), true),
                new LayerAction(new BMNGOneImage(), true),
                new LayerAction(new BMNGSurfaceLayer(), true),
                new LayerAction(new LandsatI3(), true),
                new LayerAction(new USGSDigitalOrtho(), false),
                new LayerAction(new USGSUrbanAreaOrtho(), false),
//                new LayerAction(new OpenStreetMapLayer(), false),
    //            new LayerAction(new WWXMLLayer("WW_Images.xml", "USGS Topo Maps"), false),
    //            new LayerAction(new WWXMLLayer("WW_Images.xml", "Geocover 1990"), false),
    //            new LayerAction(new WWXMLLayer("WW_Images.xml", "Geocover 2000"), false),
                new LayerAction(new CountryBoundariesLayer(), false)
              };
                        
            final LayerAction[] layerActions2 = new LayerAction[] 
              {
                new LayerAction(new PathLayer("Paths", getLookup()), true),
                new LayerAction(new WaypointLayer("Photos", getLookup()), true),
                new LayerAction(new EarthNASAPlaceNameLayer(), true),
                new LayerAction(new GridLayer(), false),
                new LayerAction(new ScalebarLayer(), true),
                new LayerAction(new WorldMapLayer(), true),
                new LayerAction(new CompassLayer(), true),
//                new LayerAction(new CompassLayer("it/tidalwave/geo/viewer/worldwindprovider/icons/compass.png"), true),
//                new LayerAction(new CompassLayer("/tmp/compass.png"), true),
              };

            final List<LayerAction> layerActions = new ArrayList<LayerAction>();
            layerActions.addAll(Arrays.asList(layerActions1));
            
            for (final LayerProvider layerProvider : Lookup.getDefault().lookupAll(LayerProvider.class))
              {
                layerActions.add(new LayerAction(layerProvider.createLayer(), layerProvider.isActiveByDefault()));
              }
                
            for (final TileProviderDescriptor tileProviderDescriptor : Lookup.getDefault().lookupAll(TileProviderDescriptor.class))
              {
                final LayerProvider layerProvider = new TiledImageLayerProvider(tileProviderDescriptor, lookup);
                layerActions.add(new LayerAction(layerProvider.createLayer(), layerProvider.isActiveByDefault()));
              }

            layerActions.addAll(Arrays.asList(layerActions2));
            final Model model = (Model)WorldWind.createConfigurationComponent(AVKey.MODEL_CLASS_NAME);

            for (final LayerAction layerAction : layerActions)
              {
                final JCheckBoxMenuItem jcb = new JCheckBoxMenuItem(layerAction);
                jcb.setSelected(layerAction.isSelected());
                optionsMenu.add(jcb);
                layerList.add(layerAction.layer);

                if (layerAction.layer instanceof TiledImageLayer)
                  {
                    ((TiledImageLayer)layerAction.layer).setShowImageTileOutlines(false);
                  }

                if (layerAction.layer instanceof LandsatI3)
                  {
                    ((TiledImageLayer)layerAction.layer).setDrawBoundingVolumes(false);
                  }

                if (layerAction.layer instanceof USGSDigitalOrtho)
                  {
                    ((TiledImageLayer)layerAction.layer).setDrawTileIDs(false);
                  }

//                if (layerAction.layer instanceof WaypointLayer)
//                  {
//                    worldWindowGLCanvas.addSelectListener((SelectListener)layerAction.layer);   TODO
//                  }
              }

            model.setLayers(layerList);

//            worldWindow = new WorldWindowGLCanvas();
            worldWindow.setModel(model);
            worldWindow.addPositionListener(positionListener);
            thePanel.setPreferredSize(new Dimension(0, 0));
            thePanel.add((Component)worldWindow, BorderLayout.CENTER);
            ready = true;
            logger.fine(">>>> initResources() completed");
          }
        catch (RuntimeException e)
          {
            logger.severe("Error in intializing resources: %s", e);
            logger.throwing(CLASS, "initResources()", e);
          }
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public JPopupMenu getOptionsMenu()
      {
        return optionsMenu;
      }
     
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Coordinate getCenterPosition()
      {
        final Dimension size = ((Component)worldWindow).getSize();
        return pointToCoordinates(new Point(size.width / 2, size.height / 2));
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void setCenterPosition (final @Nonnull Coordinate centerPosition)
      {
        logger.info("setCenterPosition(%s)", centerPosition);
        setCenterPosition(centerPosition, 30000.0 * 1000.0, Angle.ZERO, Angle.ZERO); // elevation for a good initial view of the earth
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void setCenterPosition (final @Nonnull Coordinate centerPosition,
                                    final @Nonnegative double zoom,
                                    final @Nonnull Angle heading,
                                    final @Nonnull Angle pitch)
      {
        logger.info("setCenterPosition(%s, %f, %s, %s)", centerPosition, zoom, heading, pitch);
//        this.centerPosition = centerPosition;
        final Position position = toPosition(centerPosition);
//        final Position viewPosition = new Position(position.getLatitude(), position.getLongitude(), height);
        logger.fine(">>>> position: %s", position);
        final Globe globe = worldWindow.getModel().getGlobe();
        final OrbitView view = (OrbitView)worldWindow.getView();
        
        if (centerAnimationDelay == 0)
          {
//            view.setLatLonAltitude(viewPosition);
            //view.setLookAtLatLon(position.getLatLon());
            view.setCenterPosition(position);
          }
        
        else
          {
            view.stopStateIterators();
            view.applyStateIterator(FlyToOrbitViewStateIterator.createPanToIterator(globe, 
//                            new LatLon(view.getLookAtLatitude(), view.getLookAtLongitude()),
//                            position.getLatLon(),  
                            view.getCenterPosition(),
                            position,
                            view.getHeading(), heading,
                            view.getPitch(), pitch,
                            view.getZoom(), zoom,
                            centerAnimationDelay));
          }
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void fitView (final @Nonnull Sector sector)
      {
        logger.info("fitView(%s)", sector);
        final double angle = sector.getMaxDelta();
        final OrbitView view = (OrbitView)worldWindow.getView();
        final Angle fov = view.getFieldOfView();
        // TODO: see http://forum.worldwindcentral.com/showthread.php?p=49433#post49433
        final double elevation = 6400 * 1000 * angle / (fov.getDegrees() / 2);
        logger.fine(">>>> fov: %f, sector delta: %s, elevation: %f", fov.getDegrees(), angle, elevation);
        setCenterPosition(sector.getCenter(), elevation, Angle.ZERO, Angle.fromDegrees(45));
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Component getRendererComponent() 
      {
        return thePanel;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnegative
    public int getMaximumZoomLevel() 
      {
        return MAX_ZOOM_LEVEL;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnegative
    public int getMinimumZoomLevel() 
      {
        return MIN_ZOOM_LEVEL;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void setZoom (final @Nonnegative int zoom)
      {
        // TODO: need a better abstraction for Zoom...
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnegative
    public int getZoom() 
      {
        return MIN_ZOOM_LEVEL;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void pan (final double x, final double y) 
      {
        // TODO: but pan() could be entirely dropped
      }
      
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Point coordinatesToPoint (final @Nonnull Coordinate coordinates)
      {
        return new Point(0, 0); // FIXME: to be implemented
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Coordinate pointToCoordinates (final @Nonnull Point point)
      {
        final View view = worldWindow.getView();
        return toCoordinate(view.computePositionFromScreenPoint(point.x, point.y));
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void activated()
      {
        final String message = NbBundle.getMessage(getClass(), "NASA.License.text");
        final String title = NbBundle.getMessage(getClass(), "NASA.License.title");
        DialogRunner.showOneShotMessage(message, title, NASAWorldWindProvider.class, "NASALicense.2");
        super.activated();
//        logger.info("activated() - clearing memory cache");
//        WorldWind.getMemoryCacheSet().clear();
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void deactivating()
      {
        logger.info("deactivating() - clearing memory cache");
        try
          {
            // See http://issues.worldwind.arc.nasa.gov/browse/WWJ-50
            // but it doesn't work
            worldWindow.getTextureCache().clear();
          }
        catch (Exception e)
          {
            logger.warning(e.toString());
            logger.throwing(CLASS, "deactivating()", e);
          }

        super.deactivating();
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
//    private static LatLon toLatLon (final Coordinates coordinates)
//      {
//        return new LatLon(Angle.fromDegreesLatitude(coordinates.getLatitude()),
//                          Angle.fromDegreesLongitude(coordinates.getLongitude())); 
//      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private static Coordinate toCoordinate (final @Nonnull Position position)
      {
        return new Coordinate(position.getLatitude().getDegrees(),
                              position.getLongitude().getDegrees(),
                              position.getElevation()); // FIXME: meters/feet?
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull 
    private static Position toPosition (final @Nonnull Coordinate coordinate)
      {
        return new Position(Angle.fromDegreesLatitude(coordinate.getLatitude()),
                            Angle.fromDegreesLongitude(coordinate.getLongitude()),
                            coordinate.getAltitude()); // FIXME: meters/feet?
      }
  }
