/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.viewer.worldwindprovider.impl.layer;

import java.util.logging.Logger;
import java.io.IOException;
import java.io.InputStream;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;
import gov.nasa.worldwind.WorldWind;
import gov.nasa.worldwind.avlist.AVKey;
import gov.nasa.worldwind.avlist.AVList;
import gov.nasa.worldwind.avlist.AVListImpl;
import gov.nasa.worldwind.exception.WWRuntimeException;
import gov.nasa.worldwind.geom.Angle;
import gov.nasa.worldwind.geom.LatLon;
import gov.nasa.worldwind.geom.Sector;
import gov.nasa.worldwind.layers.TiledImageLayer;
import gov.nasa.worldwind.layers.TextureTile;
import gov.nasa.worldwind.render.DrawContext;
import gov.nasa.worldwind.util.LevelSet;

/***********************************************************************************************************************
 *
 * @author Patrick Murris
 * @version $Id$
 *
 **********************************************************************************************************************/
/**
 * @version 0.1 - 20070523 - Very first try at a XML images layer loader
 * Reads a XML WW layer definition and create the corresponding TiledImagelayer
 * Issues : only support image layers from a ImageTileService node
 * No WMS support and few error handling...
 */

public class WWXMLLayer  extends TiledImageLayer 
  {
    private final static String CLASS = WWXMLLayer.class.getName();
    
    private final static Logger logger = Logger.getLogger(CLASS);
    
    public WWXMLLayer(String fileName, String layerName)
    {
        super(makeLevels(readXML(fileName), layerName));
        this.setName(layerName);
        this.setUseTransparentTextures(true);        
    }

    // Read an XML document from a file
    private static Document readXML(String fileName)
    {
        if (fileName == null)
        {
            String message = WorldWind.getStringValue("nullValue.FileIsNull");
            logger.fine(message);
            throw new IllegalArgumentException(message);
        }
        
        InputStream is = Thread.currentThread().getContextClassLoader().getResourceAsStream(WWXMLLayer.class.getPackage().getName().replace('.', '/') + "/" + fileName);

        try
        {
            DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
            docBuilderFactory.setNamespaceAware(false);
            DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
            Document doc = docBuilder.parse(is);
            return doc;
        }
        // TODO: use proper error message strings - not GeoRSS
        catch (ParserConfigurationException e)
        {
            String message = WorldWind.getStringValue("GeoRSS.ParserConfigurationException");
            logger.throwing(CLASS, message, e);
            throw new WWRuntimeException(message, e);
        }
        catch (IOException e)
        {
            String message = WorldWind.getStringValue("GeoRSS.IOExceptionParsing") + fileName;
            logger.throwing(CLASS, message, e);
            throw new WWRuntimeException(message, e);
        }
        catch (SAXException e)
        {
            String message = WorldWind.getStringValue("GeoRSS.IOExceptionParsing") + fileName;
            logger.throwing(CLASS, message, e);
            throw new WWRuntimeException(message, e);
        }
    }
    
    // Create LevelSet from xml doc for the specified layerName
    private static LevelSet makeLevels(Document doc, String layerName)
    {
    	// Get QuadtileSet node
        Node node = findQuadtileSetByName(doc, layerName);
        if(node == null)
        	return null; // TODO: throw exception

        // Extract useful info
        Node child;
        // Sector
        Sector sector = Sector.FULL_SPHERE;
        child = findChildByName(node, "BoundingBox");
        if(child != null) {
        	Node n;
          	n = findChildByName(child, "North");
        	double north = Double.parseDouble(findChildByName(n, "Value").getTextContent());
          	n = findChildByName(child, "South");
        	double south = Double.parseDouble(findChildByName(n, "Value").getTextContent());
          	n = findChildByName(child, "West");
        	double west = Double.parseDouble(findChildByName(n, "Value").getTextContent());
          	n = findChildByName(child, "East");
        	double east = Double.parseDouble(findChildByName(n, "Value").getTextContent());
        	sector = new Sector(Angle.fromDegrees(south), Angle.fromDegrees(north), Angle.fromDegrees(west), Angle.fromDegrees(east));
        }
        // Images and server access
        double levelZeroTileSizeDegree;
        int numLevels;
        int tileSize = 512;
        String service;
        String datasetName;
        String formatSuffix;
        child = findChildByName(node, "ImageAccessor");
        if(child != null) {
        	Node n;
          	n = findChildByName(child, "LevelZeroTileSizeDegrees");
          	levelZeroTileSizeDegree = Double.parseDouble(n.getTextContent());
          	n = findChildByName(child, "NumberLevels");
          	numLevels = Integer.parseInt(n.getTextContent());
          	n = findChildByName(child, "TextureSizePixels");
          	tileSize = Integer.parseInt(n.getTextContent());
          	n = findChildByName(child, "ImageFileExtension");
          	formatSuffix = "." + n.getTextContent();
          	n = findChildByName(child, "ImageTileService");
          	if(n != null) {
          		Node n2;
              	n2 = findChildByName(n, "ServerUrl");
              	service = n2.getTextContent();
              	n2 = findChildByName(n, "DataSetName");
              	datasetName = n2.getTextContent();
          	} else {
          		// No ImageTileService node - may be using wms ?
          		// TODO: throw exception
          		return null;
          	}
        } else {
        	// No Image accessor node
        	// TODO: throw exception
        	return null;
        }
        
        // Compose level param
        AVList params = new AVListImpl();
        params.setValue(AVKey.TILE_WIDTH, tileSize);
        params.setValue(AVKey.TILE_HEIGHT, tileSize);
        params.setValue(AVKey.DATA_CACHE_NAME, "Earth/" + layerName);
        params.setValue(AVKey.SERVICE, service);
        params.setValue(AVKey.DATASET_NAME, datasetName);
        params.setValue(AVKey.FORMAT_SUFFIX, formatSuffix);
        params.setValue(AVKey.NUM_LEVELS, numLevels);
        params.setValue(AVKey.NUM_EMPTY_LEVELS, 0);
        params.setValue(AVKey.LEVEL_ZERO_TILE_DELTA, new LatLon(Angle.fromDegrees(levelZeroTileSizeDegree), Angle.fromDegrees(levelZeroTileSizeDegree)));
        params.setValue(AVKey.SECTOR, sector);

        // Return LevelSet
        return new LevelSet(params);
    }

    @Override
    public String toString()
    {
        return "WWXMLLayer";
    }

    // DOM tools
 
    // Find a QuadtileSet node by its name - exact match case sens.
    private static Node findQuadtileSetByName(Document doc, String layerName) {
        NodeList nodes = doc.getElementsByTagName("QuadTileSet");
        if(nodes.getLength() < 1)
        	return null;
        // Find quadtileset node for layerName
        for (int i = 0; i < nodes.getLength(); i++)
        {
            Node node = nodes.item(i);
            Node name = findChildByName(node, "Name"); 
            if(name != null && name.getTextContent().equals(layerName))
            	return node;
        }
        return null;
    }
 
    // Find a child node by its name - not case sensitive
    private static Node findChildByName(Node parent, String name)
    {
        NodeList children = parent.getChildNodes();
        if (children == null || children.getLength() < 1)
            return null;

        for (int i = 0; i < children.getLength(); i++)
        {
            String n = children.item(i).getNodeName();
            if (n != null && n.toLowerCase().equals(name.toLowerCase()))
                return children.item(i);
        }

        return null;
    }

    @Override
    protected void requestTexture(DrawContext arg0, TextureTile arg1) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    protected void forceTextureLoad(TextureTile arg0) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

}
