/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * Copyright (C) 2007-2012 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://forceten.tidalwave.it
 * SCM: https://bitbucket.org/tidalwave/forceten-src
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.viewer.worldwindprovider.impl.layer;

import it.tidalwave.geo.Coordinate;
import java.util.List;
import javax.annotation.Nonnull;
import java.lang.reflect.InvocationTargetException;
import java.util.logging.Logger;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;
import javax.swing.SwingUtilities;
import org.jdesktop.observablecollections.ObservableList;
import org.openide.util.Lookup;
import gov.nasa.worldwind.Disposable;
import gov.nasa.worldwind.event.SelectEvent;
import gov.nasa.worldwind.event.SelectListener;
import gov.nasa.worldwind.geom.Angle;
import gov.nasa.worldwind.geom.Position;
import gov.nasa.worldwind.layers.IconLayer;
import gov.nasa.worldwind.render.UserFacingIcon;
import it.tidalwave.geo.viewer.FeatureManager;
import it.tidalwave.geo.viewer.spi.feature.FeatureManagerSpi;
import it.tidalwave.geo.viewer.spi.feature.FeaturesRenderer;
import org.jdesktop.observablecollections.ObservableListListener;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class WaypointLayer extends IconLayer implements SelectListener, Disposable // FIXME: FeatureLayer
  {
    private final static String CLASS = WaypointLayer.class.getName();
    private final static Logger logger = Logger.getLogger(CLASS);
    
    private UserFacingIcon lastToolTipIcon = null;
    
    private UserFacingIcon lastPickedIcon;

    private final ObservableListListener featureListener = new ObservableListListener()
      {
        @Override
        public void listElementsAdded (ObservableList features, int from, int count)
          {
            for (int i = from; i < from + count; i++)
              {
                addFeature(features.get(i));
              }
          }

        @Override
        public void listElementsRemoved (ObservableList ol, int i, List list)
          {
//            for (final WWIcon icon : getIcons())
//              {
//                removeIcon(icon);
//              }
          }

        @Override
        public void listElementReplaced (ObservableList ol, int i, Object o)
          {
          }

        @Override
        public void listElementPropertyChanged (ObservableList ol, int i)
          {
          }
      };

    @Nonnull
    private final Lookup lookup;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public WaypointLayer (final @Nonnull String layerName,
                          final @Nonnull Lookup lookup)
      {
        this.setName(layerName);
        this.lookup = lookup;
        final FeatureManager featureManager = lookup.lookup(FeatureManager.class);
        featureManager.getFeatures().addObservableListListener(featureListener);
      }
    
    // Do not detach the listener on dispose() - see http://forum.worldwindcentral.com/showthread.php?p=49436
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void addFeature (final @Nonnull Object feature)
      {
        // Better safe than sorry: see http://forum.worldwindcentral.com/showthread.php?t=11407
        if (SwingUtilities.isEventDispatchThread())
          {
            addFeatureInAWTThread(feature);
          }
        else
          {
            try
              {
                SwingUtilities.invokeAndWait(new Runnable()
                  {
                    @Override
                    public void run()
                      {
                        addFeatureInAWTThread(feature);
                      }
                  });
              } 
            catch (InterruptedException e)
              {
                logger.throwing(CLASS, "createSegments()", e);
              } 
            catch (InvocationTargetException e)
              {
                logger.throwing(CLASS, "createSegments()", e);
              }
          }
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void addFeatureInAWTThread (final @Nonnull Object feature)
      {
        logger.info("addFeatureInAWTThread(" + feature + ")");
        final FeatureManagerSpi featureManagerSpi = lookup.lookup(FeatureManagerSpi.class);
        final Coordinate coordinate = featureManagerSpi.findCoordinate(feature);
        final double latitude = coordinate.getLatitude();
        final double longitude = coordinate.getLongitude();
        final double distanceAboveSurface = 0; // FIXME
        final Dimension size = new Dimension();
        /*final*/ BufferedImage image = lookup.lookup(FeaturesRenderer.class).makeIcon(feature, true, size);
        //
        // On MAC OS X, any icon whose size is not a power of two causes severe flickering.
        // http://forum.worldwindcentral.com/showthread.php?p=46115#post46115
        //
        if (System.getProperty("os.name").toLowerCase().indexOf("mac os x") >= 0)
          {
            image = growToPowerOfTwo(size, image);
          }
        
        final Position position = new Position(Angle.fromDegrees(latitude), Angle.fromDegrees(longitude), distanceAboveSurface);        
        final UserFacingIcon icon = new UserFacingIcon(image, position);
        icon.setHighlightScale(1.5);
//        icon.setToolTipFont(this.makeToolTipFont());
        final String name = "theName";
        icon.setToolTipText(name);
        icon.setToolTipTextColor(Color.DARK_GRAY);
        icon.setSize(size);
        icon.setVisible(true);
        addIcon(icon);       
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void selected (final @Nonnull SelectEvent event)
      {
        if (event.getEventAction().equals(SelectEvent.LEFT_CLICK))
          {
            if (event.hasObjects() && (event.getTopObject() instanceof UserFacingIcon))
              {
                UserFacingIcon icon = (UserFacingIcon)event.getTopObject();
                // TODO: do something
              }
          }
        
        else if (event.getEventAction().equals(SelectEvent.HOVER))
          {
            if (lastToolTipIcon != null)
              {
                lastToolTipIcon.setShowToolTip(false);
                this.lastToolTipIcon = null;
                //this.wwd.repaint();
              }

            if (event.hasObjects() && (event.getTopObject() instanceof UserFacingIcon))
              {
                this.lastToolTipIcon = (UserFacingIcon) event.getTopObject();
                lastToolTipIcon.setShowToolTip(true);
                //this.wwd.repaint();
              }
          }
        
        else if (event.getEventAction().equals(SelectEvent.ROLLOVER))
          {
            this.highlight(event.getTopObject());
            //this.wwd.repaint();
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void highlight (final @Nonnull Object object)
      {
        if (this.lastPickedIcon == object)
          {
            return; // same thing selected
          }

        if (this.lastPickedIcon != null)
          {
            this.lastPickedIcon.setHighlighted(false);
            this.lastPickedIcon = null;
          }

        if (object != null && object instanceof UserFacingIcon)
          {
            this.lastPickedIcon = (UserFacingIcon)object;
            this.lastPickedIcon.setHighlighted(true);
          }
      } 
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull 
    private BufferedImage growToPowerOfTwo (final @Nonnull Dimension size, @Nonnull BufferedImage image)
      {
        int newSize = 2;
        
        while (newSize < Math.max(size.getWidth(), size.getHeight()))
          {
            newSize *= 2;  
          }

        size.setSize(newSize, newSize);
        BufferedImage origImage = image;
        image = new BufferedImage((int)size.getWidth(), (int)size.getHeight(), BufferedImage.TYPE_INT_ARGB);
        final Graphics2D g = image.createGraphics();

        try
          {
            g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
            g.drawImage(origImage, (newSize - origImage.getWidth()) / 2, newSize - origImage.getHeight(), null);
          }
        finally
          {
            g.dispose();
          }
        
        return image;
     }
  }
