/*
Copyright (C) 2001, 2006 United States Government
as represented by the Administrator of the
National Aeronautics and Space Administration.
All Rights Reserved.
  
Modified by Fabrizio Giudici.
 */
package it.tidalwave.geo.viewer.worldwindprovider.spi;

import it.tidalwave.geo.viewer.worldwindprovider.impl.MercatorSector;
import java.net.URISyntaxException;
import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.inject.Inject;
import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.awt.image.BufferedImage;
import javax.imageio.ImageIO;
import com.sun.opengl.util.texture.TextureData;
import com.sun.opengl.util.texture.TextureIO;
import gov.nasa.worldwind.Configuration;
import gov.nasa.worldwind.WorldWind;
import gov.nasa.worldwind.avlist.AVKey;
import gov.nasa.worldwind.avlist.AVList;
import gov.nasa.worldwind.avlist.AVListImpl;
import gov.nasa.worldwind.cache.BasicMemoryCache;
import gov.nasa.worldwind.cache.MemoryCache;
import gov.nasa.worldwind.formats.dds.DDSConverter;
import gov.nasa.worldwind.geom.Angle;
import gov.nasa.worldwind.geom.Vec4;
import gov.nasa.worldwind.render.DrawContext;
import gov.nasa.worldwind.retrieve.HTTPRetriever;
import gov.nasa.worldwind.retrieve.RetrievalPostProcessor;
import gov.nasa.worldwind.retrieve.Retriever;
import gov.nasa.worldwind.retrieve.URLRetriever;
import gov.nasa.worldwind.util.LevelSet;
import gov.nasa.worldwind.util.Logging;
import org.openide.util.Parameters;
import it.tidalwave.geo.viewer.spi.ResourceCache;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.util.logging.Logger;
import java.io.IOException;

/**
 * @author tag
 * @version $Id$
 * Modified 2009-02-03 to add support for Mercator projections
 */
public class BasicMercatorTiledImageLayer extends MercatorTiledImageLayer
  {
    private static final String CLASS = BasicMercatorTiledImageLayer.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);

    @Inject
    protected ResourceCache resourceCache;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static class DownloadPostProcessor implements RetrievalPostProcessor
      {
        // TODO: Rewrite this inner class, factoring out the generic parts.

        private final MercatorTextureTile tile;
        private final BasicMercatorTiledImageLayer layer;

        public DownloadPostProcessor (final @Nonnull MercatorTextureTile tile,
                                      final @Nonnull BasicMercatorTiledImageLayer layer)
          {
            this.tile = tile;
            this.layer = layer;
          }

        @Override
        @edu.umd.cs.findbugs.annotations.SuppressWarnings("REC_CATCH_EXCEPTION")
        public ByteBuffer run (final @Nonnull Retriever retriever)
          {
            Parameters.notNull("missing retriever", retriever);

            try
              {
                if (!retriever.getState().equals(Retriever.RETRIEVER_STATE_SUCCESSFUL))
                  {
                    return null;
                  }

                final URLRetriever r = (URLRetriever)retriever;
                ByteBuffer buffer = r.getBuffer();

                if (retriever instanceof HTTPRetriever)
                  {
                    final HTTPRetriever htr = (HTTPRetriever)retriever;

                    if (htr.getResponseCode() == HttpURLConnection.HTTP_NO_CONTENT)
                      {
                        // Mark tile as missing to avoid excessive attempts
                        this.layer.getLevels().markResourceAbsent(this.tile);
                        return null;
                      }
                    else if (htr.getResponseCode() != HttpURLConnection.HTTP_OK)
                      {
                        // Also mark tile as missing, but for an unknown reason.
                        this.layer.getLevels().markResourceAbsent(this.tile);
                        return null;
                      }
                  }

//                final File outFile = CACHE.newFile(this.tile.getPath());
//
//                if (outFile == null)
//                  {
//                    return null;
//                  }
//
//                if (outFile.exists())
//                  {
//                    return buffer;
//                  }

                // TODO: Better, more generic and flexible handling of file-format type
                if (buffer != null)
                  {
                    final String contentType = r.getContentType();

                    if (contentType == null)
                      {
                        // TODO: logger message
                        return null;
                      }

                    if (contentType.contains("xml") || contentType.contains("html") || contentType.contains("text"))
                      {
                        this.layer.getLevels().markResourceAbsent(this.tile);

                        StringBuffer sb = new StringBuffer();

                        while (buffer.hasRemaining())
                          {
                            sb.append((char) buffer.get());
                          }

                        // TODO: parse out the message if the content is xml or html.
                        logger.severe(sb.toString());

                        return null;
                      }
                    else if (contentType.contains("dds"))
                      {
                        saveBuffer(retriever);
                      }
                    else if (contentType.contains("zip"))
                      {
                        saveBuffer(retriever);
                        // Assume it's zipped DDS, which the retriever would have unzipped into the buffer.
//                        this.layer.saveBuffer(buffer, outFile);
                      }
//                    else if (outFile.getName().endsWith(".dds"))
                    else if (tile.getPath().endsWith(".dds"))
                      {
                        // Convert to DDS and save the result.
                        buffer = DDSConverter.convertToDDS(buffer, contentType);

                        if (buffer != null)
                          {
                            saveBuffer(retriever);
                          }
                      }
                    else if (contentType.contains("image"))
//                             && !this.layer.transformImage(buffer,
//                                                           tile.getMercatorSector(),
//                                                           tile.getResourceURL().toExternalForm()))
                      {
                        saveBuffer(retriever);
                      }

                    if (buffer != null)
                      {
                        this.layer.firePropertyChange(AVKey.LAYER, null, this);
                      }

                    return buffer;
                  }
              }
            catch (Exception e)
              {
                this.layer.getLevels().markResourceAbsent(this.tile);
                logger.severe("Exception saving texture file: %s, %s", tile.getPath(), e);
                logger.throwing(CLASS, CLASS, e);
              }

            return null;
          }

        private void saveBuffer (final @Nonnull Retriever retriever)
          throws MalformedURLException, URISyntaxException
          {
            layer.resourceCache.storeImage(tile.getResourceURL().toURI(), retriever.getBuffer().array());
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static class RequestTask implements Runnable, Comparable<RequestTask>
      {
        private final BasicMercatorTiledImageLayer layer;
        private final MercatorTextureTile tile;

        private RequestTask (final @Nonnull MercatorTextureTile tile,
                             final @Nonnull BasicMercatorTiledImageLayer layer)
          {
            this.layer = layer;
            this.tile = tile;
          }

        @Override
        public void run()
          {
            try
              {
                final URL textureURL = layer.resourceCache.findCachedResourceUri(tile.getResourceURL().toURI()).toURL();

//                if (textureURL != null && !this.layer.isTextureExpired(tile, textureURL))
                if (this.layer.loadTexture(tile, textureURL))
                  {
                    layer.getLevels().unmarkResourceAbsent(tile);
                    this.layer.firePropertyChange(AVKey.LAYER, null, this);
                  }
                else
                  {
//                    CACHE.removeFile(textureURL);
                    layer.getLevels().markResourceAbsent(tile);
                    logger.warning("Corrupted texture: %s", textureURL);
                  }
              } 
            catch (NotFoundException e)
              {
                layer.downloadTexture(tile);
              }
            catch (Exception e)
              {
                throw new RuntimeException(e);
              }
          }

        @Override
        public int compareTo (final @Nonnull RequestTask that)
          {
            Parameters.notNull("null that", that);

            return this.tile.getPriority() == that.tile.getPriority() ? 0
                    : this.tile.getPriority() < that.tile.getPriority() ? -1
                    : 1;
          }

        @Override
        public boolean equals (final @CheckForNull Object o)
          {
            if (this == o)
              {
                return true;
              }

            if (o == null || getClass() != o.getClass())
              {
                return false;
              }

            final RequestTask that = (RequestTask) o;

            // Don't include layer in comparison so that requests are shared among layers
            return !(tile != null ? !tile.equals(that.tile) : that.tile != null);
          }

        @Override
        public int hashCode()
          {
            return (tile != null ? tile.hashCode() : 0);
          }

        @Override @Nonnull
        public String toString()
          {
            return tile.toString();
          }
      }

    private final Object fileLock = new Object();

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public BasicMercatorTiledImageLayer (final @Nonnull LevelSet levelSet, final @Nonnull String name)
      {
        super(levelSet, name);

        if (!WorldWind.getMemoryCacheSet().containsCache(MercatorTextureTile.class.getName()))
          {
            final long size = Configuration.getLongValue(AVKey.TEXTURE_IMAGE_CACHE_SIZE, 3000000L);
            final MemoryCache cache = new BasicMemoryCache((long)(0.85 * size), size);
            cache.setName("Texture Tiles");
            WorldWind.getMemoryCacheSet().addCache(MercatorTextureTile.class.getName(), cache);
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public BasicMercatorTiledImageLayer (final @Nonnull AVList params, final @Nonnull String name)
      {
        this(new LevelSet(params), name);
        this.setValue(AVKey.CONSTRUCTION_PARAMETERS, params);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    protected void forceTextureLoad (final @Nonnull MercatorTextureTile tile)
      {
        logger.finer("forceTextureLoad(%s)", tile);
        
        try
          {
//                CACHE.findFile(tile.getPath(), true);
            final URL textureURL = resourceCache.findCachedResourceUri(tile.getResourceURL().toURI()).toURL();
//        if (textureURL != null && !this.isTextureExpired(tile, textureURL))
            this.loadTexture(tile, textureURL);
          }
        catch (NotFoundException e)
          {
            // FIXME: ??
          }
        catch (Exception e)
          {
            throw new RuntimeException(e);
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    protected void requestTexture (final @Nonnull DrawContext dc, final @Nonnull MercatorTextureTile tile)
      {
        logger.finer("requestTexture(%s, %s)", dc, tile);
        
        final Vec4 centroid = tile.getCentroidPoint(dc.getGlobe());

        if (this.getReferencePoint() != null)
          {
            tile.setPriority(centroid.distanceTo3(this.getReferencePoint()));
          }

        final RequestTask task = new RequestTask(tile, this);
        this.getRequestQ().add(task);
      }

//    /*******************************************************************************************************************
//     *
//     *
//     ******************************************************************************************************************/
//    private boolean isTextureExpired (final @Nonnull MercatorTextureTile tile, final @Nonnull URL textureURL)
//      {
//        if (!WWIO.isFileOutOfDate(textureURL, tile.getLevel().getExpiryTime()))
//          {
//            return false;
//          }
//
//        // The file has expired. Delete it.
////        CACHE.removeFile(textureURL);
////        String message = Logging.getMessage("generic.DataFileExpired", textureURL);
////        logger.fine(message);
//        return true;
//      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private boolean loadTexture (final @Nonnull MercatorTextureTile tile, final @Nonnull URL textureURL)
      {
        logger.finer("loadTexture(%s, %s)", tile, textureURL);
        
        TextureData textureData;

        synchronized (this.fileLock)
          {
            textureData = readTexture(textureURL, this.isUseMipMaps());
          }

        if (textureData == null)
          {
            return false;
          }

        tile.setTextureData(textureData);

        if (tile.getLevelNumber() != 0 || !this.isRetainLevelZeroTiles())
          {
            this.addTileToCache(tile);
          }

        return true;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static TextureData readTexture (final @Nonnull URL url, boolean useMipMaps)
      {
        logger.finer("readTexture(%s, %s)", url, useMipMaps);

        try
          {
            return TextureIO.newTextureData(url, useMipMaps, null);
          }
        catch (Exception e)
          {
            logger.severe("Exception reading texture: %s %s", url, e);
            logger.throwing(CLASS, "readTexture", e);

            return null;
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void addTileToCache (final @Nonnull MercatorTextureTile tile)
      {
        WorldWind.getMemoryCache(MercatorTextureTile.class.getName()).add(tile.getTileKey(), tile);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void downloadTexture (final @Nonnull MercatorTextureTile tile)
      {
        logger.finer("downloadTexture(%s)", tile);

        if (!WorldWind.getRetrievalService().isAvailable())
          {
            return;
          }

        URL url = null;
        
        try
          {
            url = tile.getResourceURL();
            if (url == null)
              {
                return;
              }

            if (WorldWind.getNetworkStatus().isHostUnavailable(url))
              {
                return;
              }
          }
        catch (MalformedURLException e)
          {
            logger.severe("Exception creating texture: %s %s", tile, e);
            logger.throwing(CLASS, CLASS, e);
            return;
          }

        Retriever retriever;

        if ("http".equalsIgnoreCase(url.getProtocol()))
          {
            retriever = new HTTPRetriever(url, new DownloadPostProcessor(tile, this));
          }
        else
          {
            logger.severe(
                    Logging.getMessage(
                    "layers.TextureLayer.UnknownRetrievalProtocol", url.toString()));
            return;
          }

        // Apply any overridden timeouts.
        Integer cto = AVListImpl.getIntegerValue(this, AVKey.URL_CONNECT_TIMEOUT);

        if (cto != null && cto > 0)
          {
            retriever.setConnectTimeout(cto);
          }

        Integer cro = AVListImpl.getIntegerValue(this, AVKey.URL_READ_TIMEOUT);

        if (cro != null && cro > 0)
          {
            retriever.setReadTimeout(cro);
          }

        Integer srl = AVListImpl.getIntegerValue(this, AVKey.RETRIEVAL_QUEUE_STALE_REQUEST_LIMIT);

        if (srl != null && srl > 0)
          {
            retriever.setStaleRequestLimit(srl);
          }

        WorldWind.getRetrievalService().runRetriever(retriever, tile.getPriority());
      }

//    /*******************************************************************************************************************
//     *
//     *
//     ******************************************************************************************************************/
//    private void saveBuffer (final @Nonnull ByteBuffer buffer, final @Nonnull File outFile)
//      throws IOException
//      {
//        logger.finer("saveBuffer(%s, %s)", buffer, outFile);
//
//        synchronized (this.fileLock)
//          {
//            WWIO.saveBuffer(buffer, outFile);
//          }
//      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private boolean transformImage (final @Nonnull ByteBuffer buffer,
                                    final @Nonnull MercatorSector sector,
                                    final @Nonnull String path)
      {
        logger.finer("transformImage(%s, %s, %s)", buffer, sector, path);
        try
          {
            final InputStream is = new ByteArrayInputStream(buffer.array());
            BufferedImage image = ImageIO.read(is);
            // FIXME: the transformed image is not used. Transformed images are "stretched" to compensate
            // mercator distortion at high latitudes, and I see some errors in the renderer, so perhaps sometimes
            // the cached, untrasformed tile is used?
            // Measure how long transform takes and eventually apply it on the fly after reading from the cache
            image = transform(image, sector);

            resourceCache.storeImage(new URI(path), buffer.array());
            return true;
          }
        catch (URISyntaxException e)
          {
            return false;
          }
        catch (IOException e)
          {
            return false;
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private BufferedImage transform (final @Nonnull BufferedImage image,
                                     final @Nonnull MercatorSector sector)
      {
        logger.finer("transform(%s, %s)", image, sector);
        final long time = System.currentTimeMillis();

        final BufferedImage transformedImage = new BufferedImage(image.getWidth(), image.getHeight(), image.getType());
        final double miny = sector.getMinLatPercent();
        final double maxy = sector.getMaxLatPercent();

        for (int y = 0; y < image.getHeight(); y++)
          {
            final double sy = 1.0 - y / (double) (image.getHeight() - 1);
            final Angle lat = Angle.fromRadians(sy * sector.getDeltaLatRadians() + sector.getMinLatitude().radians);
            double dy = 1.0 - (MercatorSector.gudermannianInverse(lat) - miny) / (maxy - miny);
            dy = Math.max(0.0, Math.min(1.0, dy));
            final int iy = (int)(dy * (image.getHeight() - 1));

            for (int x = 0; x < image.getWidth(); x++)
              {
                transformedImage.setRGB(x, y, image.getRGB(x, iy));
              }
          }

        logger.finest(">>>> image transformed in %d msec.", System.currentTimeMillis() - time);

        return transformedImage;
      }
  }
